# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for jraph.utils."""

import functools
import os
from absl.testing import absltest
from absl.testing import parameterized
import jax
from jax.lib import xla_bridge
import jax.numpy as jnp
import jax.tree_util as tree
from jraph._src import graph
from jraph._src import utils
import numpy as np


def _get_random_graph(max_n_graph=10,
                      include_node_features=True,
                      include_edge_features=True,
                      include_globals=True):
  n_graph = np.random.randint(1, max_n_graph + 1)
  n_node = np.random.randint(0, 10, n_graph)
  n_edge = np.random.randint(0, 20, n_graph)
  # We cannot have any edges if there are no nodes.
  n_edge[n_node == 0] = 0

  senders = []
  receivers = []
  offset = 0
  for n_node_in_graph, n_edge_in_graph in zip(n_node, n_edge):
    if n_edge_in_graph != 0:
      senders += list(
          np.random.randint(0, n_node_in_graph, n_edge_in_graph) + offset)
      receivers += list(
          np.random.randint(0, n_node_in_graph, n_edge_in_graph) + offset)
    offset += n_node_in_graph
  if include_globals:
    global_features = jnp.asarray(np.random.random(size=(n_graph, 5)))
  else:
    global_features = None
  if include_node_features:
    nodes = jnp.asarray(np.random.random(size=(np.sum(n_node), 4)))
  else:
    nodes = None

  if include_edge_features:
    edges = jnp.asarray(np.random.random(size=(np.sum(n_edge), 3)))
  else:
    edges = None
  return graph.GraphsTuple(
      n_node=jnp.asarray(n_node),
      n_edge=jnp.asarray(n_edge),
      nodes=nodes,
      edges=edges,
      globals=global_features,
      senders=jnp.asarray(senders),
      receivers=jnp.asarray(receivers))


def _make_nest(array):
  """Returns a nest given an array."""
  return {'a': array,
          'b': [jnp.ones_like(array), {'c': jnp.zeros_like(array)}]}


def _get_list_and_batched_graph():
  """Returns a list of individual graphs and a batched version.

  This test-case includes the following corner-cases:
    - single node,
    - multiple nodes,
    - no edges,
    - single edge,
    - and multiple edges.
  """
  batched_graph = graph.GraphsTuple(
      n_node=jnp.array([1, 3, 1, 0, 2, 0, 0]),
      n_edge=jnp.array([2, 5, 0, 0, 1, 0, 0]),
      nodes=_make_nest(jnp.arange(14).reshape(7, 2)),
      edges=_make_nest(jnp.arange(24).reshape(8, 3)),
      globals=_make_nest(jnp.arange(14).reshape(7, 2)),
      senders=jnp.array([0, 0, 1, 1, 2, 3, 3, 6]),
      receivers=jnp.array([0, 0, 2, 1, 3, 2, 1, 5]))

  list_graphs = [
      graph.GraphsTuple(
          n_node=jnp.array([1]),
          n_edge=jnp.array([2]),
          nodes=_make_nest(jnp.array([[0, 1]])),
          edges=_make_nest(jnp.array([[0, 1, 2], [3, 4, 5]])),
          globals=_make_nest(jnp.array([[0, 1]])),
          senders=jnp.array([0, 0]),
          receivers=jnp.array([0, 0])),
      graph.GraphsTuple(
          n_node=jnp.array([3]),
          n_edge=jnp.array([5]),
          nodes=_make_nest(jnp.array([[2, 3], [4, 5], [6, 7]])),
          edges=_make_nest(
              jnp.array([[6, 7, 8], [9, 10, 11], [12, 13, 14], [15, 16, 17],
                         [18, 19, 20]])),
          globals=_make_nest(jnp.array([[2, 3]])),
          senders=jnp.array([0, 0, 1, 2, 2]),
          receivers=jnp.array([1, 0, 2, 1, 0])),
      graph.GraphsTuple(
          n_node=jnp.array([1]),
          n_edge=jnp.array([0]),
          nodes=_make_nest(jnp.array([[8, 9]])),
          edges=_make_nest(jnp.zeros((0, 3))),
          globals=_make_nest(jnp.array([[4, 5]])),
          senders=jnp.array([]),
          receivers=jnp.array([])),
      graph.GraphsTuple(
          n_node=jnp.array([0]),
          n_edge=jnp.array([0]),
          nodes=_make_nest(jnp.zeros((0, 2))),
          edges=_make_nest(jnp.zeros((0, 3))),
          globals=_make_nest(jnp.array([[6, 7]])),
          senders=jnp.array([]),
          receivers=jnp.array([])),
      graph.GraphsTuple(
          n_node=jnp.array([2]),
          n_edge=jnp.array([1]),
          nodes=_make_nest(jnp.array([[10, 11], [12, 13]])),
          edges=_make_nest(jnp.array([[21, 22, 23]])),
          globals=_make_nest(jnp.array([[8, 9]])),
          senders=jnp.array([1]),
          receivers=jnp.array([0])),
      graph.GraphsTuple(
          n_node=jnp.array([0]),
          n_edge=jnp.array([0]),
          nodes=_make_nest(jnp.zeros((0, 2))),
          edges=_make_nest(jnp.zeros((0, 3))),
          globals=_make_nest(jnp.array([[10, 11]])),
          senders=jnp.array([]),
          receivers=jnp.array([])),
      graph.GraphsTuple(
          n_node=jnp.array([0]),
          n_edge=jnp.array([0]),
          nodes=_make_nest(jnp.zeros((0, 2))),
          edges=_make_nest(jnp.zeros((0, 3))),
          globals=_make_nest(jnp.array([[12, 13]])),
          senders=jnp.array([]),
          receivers=jnp.array([])),
      graph.GraphsTuple(
          n_node=jnp.array([]),
          n_edge=jnp.array([]),
          nodes=_make_nest(jnp.zeros((0, 2))),
          edges=_make_nest(jnp.zeros((0, 3))),
          globals=_make_nest(jnp.zeros((0, 2))),
          senders=jnp.array([]),
          receivers=jnp.array([])),
  ]

  return list_graphs, batched_graph


class GraphTest(parameterized.TestCase):

  def test_batch(self):
    """Tests batching of graph."""
    list_graphs_tuple, batched_graphs_tuple = _get_list_and_batched_graph()
    graphs_tuple = utils.batch(list_graphs_tuple)
    jax.tree_util.tree_map(np.testing.assert_allclose, graphs_tuple,
                           batched_graphs_tuple)

  def test_unbatch(self):
    """Tests unbatching of graph."""
    list_graphs_tuple, batched_graphs_tuple = _get_list_and_batched_graph()
    graphs_tuples = utils.unbatch(batched_graphs_tuple)
    # The final GraphsTuple does not contain a graph, and so shouldn't be
    # present in the result.
    jax.tree_util.tree_map(np.testing.assert_allclose, graphs_tuples,
                           list_graphs_tuple[:-1])

  def test_batch_np(self):
    """Tests batching of graph in numpy."""
    (list_graphs_tuple, batched_graphs_tuple) = _get_list_and_batched_graph()
    graphs_tuple = utils.batch_np(list_graphs_tuple)
    jax.tree_util.tree_map(np.testing.assert_allclose, graphs_tuple,
                           batched_graphs_tuple)

  def test_unbatch_np(self):
    """Tests unbatching of graph in numpy."""
    (list_graphs_tuple, batched_graphs_tuple) = _get_list_and_batched_graph()
    graphs_tuples = utils.unbatch_np(batched_graphs_tuple)
    # The final GraphsTuple does not contain a graph, and so shouldn't be
    # present in the result.
    jax.tree_util.tree_map(np.testing.assert_allclose, graphs_tuples,
                           list_graphs_tuple[:-1])

  @parameterized.parameters((True, True, False),
                            (True, False, True),
                            (False, True, True))
  def test_batch_unbatch_with_random_graphs(self,
                                            include_globals,
                                            include_nodes,
                                            include_edges):
    """Tests batch(unbatch) is identity with random graphs."""
    np.random.seed(42)
    for _ in range(100):
      g = _get_random_graph(include_globals=include_globals,
                            include_node_features=include_nodes,
                            include_edge_features=include_edges)
      jax.tree_util.tree_map(np.testing.assert_allclose,
                             utils.batch(utils.unbatch(g)), g)

    for _ in range(10):
      rg = lambda: _get_random_graph(  # pylint: disable=g-long-lambda
          1, include_nodes, include_edges, include_globals)
      graphs1 = [rg() for _ in range(np.random.randint(1, 10))]
      graphs2 = utils.unbatch(utils.batch(graphs1))
      for g1, g2 in zip(graphs1, graphs2):
        jax.tree_util.tree_map(np.testing.assert_allclose, g1, g2)

  def test_pad_with_graphs(self):
    """Tests padding of graph."""
    _, graphs_tuple = _get_list_and_batched_graph()
    padded_graphs_tuple = utils.pad_with_graphs(graphs_tuple, 10, 12, 9)
    expected_padded_graph = graph.GraphsTuple(
        n_node=jnp.concatenate([graphs_tuple.n_node,
                                jnp.array([3, 0])]),
        n_edge=jnp.concatenate([graphs_tuple.n_edge,
                                jnp.array([4, 0])]),
        nodes=tree.tree_map(
            lambda f: jnp.concatenate([f, jnp.zeros((3, 2), dtype=f.dtype)]),
            graphs_tuple.nodes),
        edges=tree.tree_map(
            lambda f: jnp.concatenate([f, jnp.zeros((4, 3), dtype=f.dtype)]),
            graphs_tuple.edges),
        globals=tree.tree_map(
            lambda f: jnp.concatenate([f, jnp.zeros((2, 2), dtype=f.dtype)]),
            graphs_tuple.globals),
        senders=jnp.concatenate([graphs_tuple.senders,
                                 jnp.array([7, 7, 7, 7])]),
        receivers=jnp.concatenate(
            [graphs_tuple.receivers,
             jnp.array([7, 7, 7, 7])]),
    )
    jax.tree_util.tree_map(np.testing.assert_allclose, padded_graphs_tuple,
                           expected_padded_graph)

  def test_unpad(self):
    """Tests unpadding of graph."""
    _, graphs_tuple = _get_list_and_batched_graph()
    unpadded_graphs_tuple = utils.unpad_with_graphs(graphs_tuple)
    expected_unpadded_graph = graph.GraphsTuple(
        n_node=jnp.array([1, 3, 1, 0]),
        n_edge=jnp.array([2, 5, 0, 0]),
        nodes=_make_nest(jnp.arange(10).reshape(5, 2)),
        edges=_make_nest(jnp.arange(21).reshape(7, 3)),
        globals=_make_nest(jnp.arange(8).reshape(4, 2)),
        senders=jnp.array([0, 0, 1, 1, 2, 3, 3]),
        receivers=jnp.array([0, 0, 2, 1, 3, 2, 1]))
    jax.tree_util.tree_map(np.testing.assert_allclose, unpadded_graphs_tuple,
                           expected_unpadded_graph)

  @parameterized.parameters((True, True, False),
                            (True, False, True),
                            (False, True, True))
  def test_pad_unpad_with_random_graphs(self,
                                        include_globals,
                                        include_nodes,
                                        include_edges):
    """Tests unpad(pad) is identity with random graphs."""
    np.random.seed(42)
    for _ in range(100):
      g = _get_random_graph(include_globals=include_globals,
                            include_node_features=include_nodes,
                            include_edge_features=include_edges)
      jax.tree_util.tree_map(
          np.testing.assert_allclose,
          utils.unpad_with_graphs(utils.pad_with_graphs(g, 101, 200, 11)), g)

  def test_pad_unpad_with_graphs_exact_padding(self):
    """Tests unpad(pad) is identity with random graphs."""
    g = _get_random_graph(include_globals=True,
                          include_node_features=True,
                          include_edge_features=True)
    recovered_g = utils.unpad_with_graphs(utils.pad_with_graphs(
        g,
        n_node=g.n_node.sum() + 1,
        n_edge=g.n_edge.sum(),
        n_graph=g.n_node.shape[0] + 1))

    jax.tree_util.tree_map(np.testing.assert_allclose, recovered_g, g)

  def test_get_number_of_padding_with_graphs_graphs(self):
    """Tests the number of padding graphs calculation."""
    _, graphs_tuple = _get_list_and_batched_graph()
    expected = 3
    with self.subTest('nojit'):
      jax.tree_util.tree_map(
          np.testing.assert_allclose,
          utils.get_number_of_padding_with_graphs_graphs(graphs_tuple),
          expected)
    with self.subTest('jit'):
      jax.tree_util.tree_map(
          np.testing.assert_allclose,
          jax.jit(utils.get_number_of_padding_with_graphs_graphs)(graphs_tuple),
          expected)

  def test_get_number_of_padding_with_graphs_nodes(self):
    """Tests the number of padding nodes calculation."""
    _, graphs_tuple = _get_list_and_batched_graph()
    expected = 2
    with self.subTest('nojit'):
      jax.tree_util.tree_map(
          np.testing.assert_allclose,
          utils.get_number_of_padding_with_graphs_nodes(graphs_tuple), expected)
    with self.subTest('jit'):
      jax.tree_util.tree_map(
          np.testing.assert_allclose,
          jax.jit(utils.get_number_of_padding_with_graphs_nodes)(graphs_tuple),
          expected)

  def test_get_number_of_padding_with_graphs_edges(self):
    """Tests the number of padding edges calculation."""
    _, graphs_tuple = _get_list_and_batched_graph()
    expected = 1
    with self.subTest('nojit'):
      jax.tree_util.tree_map(
          np.testing.assert_allclose,
          utils.get_number_of_padding_with_graphs_edges(graphs_tuple), expected)
    with self.subTest('jit'):
      jax.tree_util.tree_map(
          np.testing.assert_allclose,
          jax.jit(utils.get_number_of_padding_with_graphs_edges)(graphs_tuple),
          expected)

  def test_get_node_padding_mask(self):
    """Tests construction of node padding mask."""
    _, graphs_tuple = _get_list_and_batched_graph()
    expected_mask = jnp.array([1, 1, 1, 1, 1, 0, 0]).astype(bool)
    with self.subTest('nojit'):
      mask = utils.get_node_padding_mask(graphs_tuple)
      jax.tree_util.tree_map(np.testing.assert_array_equal, mask, expected_mask)
    with self.subTest('jit'):
      mask = jax.jit(utils.get_node_padding_mask)(graphs_tuple)
      jax.tree_util.tree_map(np.testing.assert_array_equal, mask, expected_mask)

  def test_get_edge_padding_mask(self):
    """Tests construction of edge padding mask."""
    _, graphs_tuple = _get_list_and_batched_graph()
    expected_mask = jnp.array([1, 1, 1, 1, 1, 1, 1, 0]).astype(bool)
    with self.subTest('nojit'):
      mask = utils.get_edge_padding_mask(graphs_tuple)
      np.testing.assert_array_equal(mask, expected_mask)
    with self.subTest('jit'):
      mask = jax.jit(utils.get_edge_padding_mask)(graphs_tuple)
      np.testing.assert_array_equal(mask, expected_mask)

  def test_get_graph_padding_mask(self):
    """Tests construction of graph padding mask."""
    _, graphs_tuple = _get_list_and_batched_graph()
    expected_mask = jnp.array([1, 1, 1, 1, 0, 0, 0]).astype(bool)
    with self.subTest('nojit'):
      mask = utils.get_graph_padding_mask(graphs_tuple)
      np.testing.assert_array_equal(mask, expected_mask)
    with self.subTest('jit'):
      mask = jax.jit(utils.get_graph_padding_mask)(graphs_tuple)
      np.testing.assert_array_equal(mask, expected_mask)

  def test_segment_sum(self):
    result = utils.segment_sum(
        jnp.arange(9), jnp.array([0, 1, 2, 0, 4, 0, 1, 1, 0]), 6)
    np.testing.assert_allclose(result, jnp.array([16, 14, 2, 0, 4, 0]))

  def test_segment_sum_optional_num_segments(self):
    result = utils.segment_sum(
        jnp.arange(9), jnp.array([0, 1, 2, 0, 4, 0, 1, 1, 0]))
    np.testing.assert_allclose(result, jnp.array([16, 14, 2, 0, 4]))

  @parameterized.parameters((True,), (False,))
  def test_segment_mean(self, nan_data):
    data = jnp.arange(9, dtype=jnp.float32)
    expected_out = jnp.array([4, 14 / 3.0, 2, 0, 4, 0])
    segment_ids = jnp.array([0, 1, 2, 0, 4, 0, 1, 1, 0])
    if nan_data:
      data = data.at[0].set(jnp.nan)
      expected_out = expected_out.at[segment_ids[0]].set(jnp.nan)
    result = utils.segment_mean(data, segment_ids, 6)
    np.testing.assert_allclose(result, expected_out)

  @parameterized.parameters((True,), (False,))
  def test_segment_variance(self, nan_data):
    data = jnp.arange(8, dtype=jnp.float32)
    expected_out = jnp.stack([jnp.var(jnp.arange(3)),
                              jnp.var(jnp.arange(3, 5)),
                              jnp.var(jnp.arange(5, 8))])
    segment_ids = jnp.array([0, 0, 0, 1, 1, 2, 2, 2])
    if nan_data:
      data = data.at[0].set(jnp.nan)
      expected_out = expected_out.at[segment_ids[0]].set(jnp.nan)
    result = utils.segment_variance(data, segment_ids, 3)
    np.testing.assert_allclose(result, expected_out)

  @parameterized.parameters((True,), (False,))
  def test_segment_normalize(self, nan_data):
    norm = lambda x: (x - jnp.mean(x)) * jax.lax.rsqrt(jnp.var(x))
    data = jnp.arange(8, dtype=jnp.float32)
    segment_ids = jnp.array([0, 0, 0, 1, 1, 2, 2, 2])
    expected_out = jnp.concatenate(
        [norm(jnp.arange(3, dtype=jnp.float32)),
         norm(jnp.arange(3, 5, dtype=jnp.float32)),
         norm(jnp.arange(5, 8, dtype=jnp.float32))])
    if nan_data:
      data = data.at[0].set(jnp.nan)
      expected_out = expected_out.at[:3].set(jnp.nan)
    result = utils.segment_normalize(data, segment_ids, 3)
    np.testing.assert_allclose(result, expected_out)

  @parameterized.parameters((False, False),
                            (True, False),
                            (True, True),
                            (False, True),
                            (False, True))
  def test_segment_max(self, indices_are_sorted, unique_indices):
    neg_inf = jnp.iinfo(jnp.int32).min
    if unique_indices:
      data = jnp.arange(6)
      if indices_are_sorted:
        segment_ids = jnp.array([0, 1, 2, 3, 4, 5])
        expected_out = jnp.array([0, 1, 2, 3, 4, 5])
        num_segments = 6
      else:
        segment_ids = jnp.array([1, 0, 2, 4, 3, -5])
        expected_out = jnp.array([1, 0, 2, 4, 3])
        num_segments = 5
    else:
      data = jnp.arange(9)
      if indices_are_sorted:
        segment_ids = jnp.array([0, 0, 0, 1, 1, 1, 2, 3, 4])
        expected_out = jnp.array([2, 5, 6, 7, 8, neg_inf])
      else:
        segment_ids = jnp.array([0, 1, 2, 0, 4, 0, 1, 1, -6])
        expected_out = jnp.array([5, 7, 2, neg_inf, 4, neg_inf])
      num_segments = 6

    with self.subTest('nojit'):
      result = utils.segment_max(data, segment_ids, num_segments,
                                 indices_are_sorted, unique_indices)
      np.testing.assert_allclose(result, expected_out)
      result = utils.segment_max(data, segment_ids,
                                 indices_are_sorted=indices_are_sorted,
                                 unique_indices=unique_indices)
      num_unique_segments = jnp.max(segment_ids) + 1
      np.testing.assert_allclose(result, expected_out[:num_unique_segments])
    with self.subTest('jit'):
      result = jax.jit(utils.segment_max, static_argnums=(2, 3, 4))(
          data, segment_ids, num_segments, indices_are_sorted, unique_indices)
      np.testing.assert_allclose(result, expected_out)

  @parameterized.parameters((False, False), (True, False),
                            (True, True), (False, True),
                            (False, True))
  def test_segment_max_or_constant(self, indices_are_sorted, unique_indices):
    if unique_indices:
      data = jnp.arange(6, dtype=jnp.float32)
      if indices_are_sorted:
        segment_ids = jnp.array([0, 1, 2, 3, 4, 5])
        expected_out = jnp.array([0, 1, 2, 3, 4, 5, 0], dtype=jnp.float32)
        num_segments = 7
      else:
        segment_ids = jnp.array([1, 0, 2, 4, 3, -5])
        expected_out = jnp.array([1, 0, 2, 4, 3], dtype=jnp.float32)
        num_segments = 5
    else:
      data = jnp.arange(9, dtype=jnp.float32)
      if indices_are_sorted:
        segment_ids = jnp.array([0, 0, 0, 1, 1, 1, 2, 3, 4])
        expected_out = jnp.array([2, 5, 6, 7, 8, 0], dtype=jnp.float32)
      else:
        segment_ids = jnp.array([0, 1, 2, 0, 4, 0, 1, 1, -6])
        expected_out = jnp.array([5, 7, 2, 0, 4, 0], dtype=jnp.float32)
      num_segments = 6

    with self.subTest('nojit'):
      result = utils.segment_max_or_constant(data, segment_ids, num_segments,
                                             indices_are_sorted, unique_indices)
      np.testing.assert_allclose(result, expected_out)
      grad = jax.grad(lambda *x: jnp.sum(utils.segment_max_or_constant(*x)))(
          data, segment_ids, num_segments, indices_are_sorted, unique_indices)
      assert np.all(jnp.isfinite(grad))
      result = utils.segment_max_or_constant(
          data,
          segment_ids,
          indices_are_sorted=indices_are_sorted,
          unique_indices=unique_indices)
      num_unique_segments = jnp.max(segment_ids) + 1
      np.testing.assert_allclose(result, expected_out[:num_unique_segments])
    with self.subTest('jit'):
      result = jax.jit(
          utils.segment_max_or_constant,
          static_argnums=(2, 3, 4))(data, segment_ids, num_segments,
                                    indices_are_sorted, unique_indices)
      np.testing.assert_allclose(result, expected_out)
      grad_fn = jax.jit(
          jax.grad(lambda *x: jnp.sum(utils.segment_max_or_constant(*x))),
          static_argnums=(2, 3, 4))
      grad = grad_fn(data, segment_ids, num_segments, indices_are_sorted,
                     unique_indices)
      assert np.all(jnp.isfinite(grad))

  @parameterized.parameters((False, False), (True, False), (True, True),
                            (False, True))
  def test_segment_max_or_constant_2d(self, indices_are_sorted, unique_indices):
    if unique_indices:
      data = jnp.stack([jnp.arange(6), jnp.arange(6, 0, -1)], axis=1)
      if indices_are_sorted:
        segment_ids = jnp.array([0, 1, 2, 3, 4, 5])
        expected_out = jnp.array(
            [[0, 6], [1, 5], [2, 4], [3, 3], [4, 2], [5, 1]])
        num_segments = 6
      else:
        segment_ids = jnp.array([1, 0, 2, 4, 3, -5])
        expected_out = jnp.array(
            [[1, 5], [0, 6], [2, 4], [4, 2], [3, 3]])
        num_segments = 5
    else:
      data = jnp.stack([jnp.arange(9), jnp.arange(9, 0, -1)], axis=1)
      if indices_are_sorted:
        segment_ids = jnp.array([0, 0, 0, 1, 1, 1, 2, 3, 4])
        expected_out = jnp.array(
            [[2, 9], [5, 6], [6, 3], [7, 2], [8, 1], [0, 0]])
      else:
        segment_ids = jnp.array([0, 1, 2, 0, 4, 0, 1, 1, -6])
        expected_out = jnp.array(
            [[5, 9], [7, 8], [2, 7], [0, 0], [4, 5], [0, 0]])
      num_segments = 6

    with self.subTest('nojit'):
      result = utils.segment_max_or_constant(data, segment_ids, num_segments,
                                             indices_are_sorted, unique_indices)
      np.testing.assert_allclose(result, expected_out)
      result = utils.segment_max_or_constant(
          data,
          segment_ids,
          indices_are_sorted=indices_are_sorted,
          unique_indices=unique_indices)
      num_unique_segments = jnp.max(segment_ids) + 1
      np.testing.assert_allclose(result, expected_out[:num_unique_segments])
    with self.subTest('jit'):
      result = jax.jit(utils.segment_max_or_constant, static_argnums=(2, 3, 4))(
          data, segment_ids, num_segments, indices_are_sorted, unique_indices)
      np.testing.assert_allclose(result, expected_out)

  @parameterized.parameters((False, False),
                            (True, False),
                            (True, True),
                            (False, True))
  def test_segment_min(self, indices_are_sorted, unique_indices):
    inf = jnp.iinfo(jnp.int32).max
    if unique_indices:
      data = jnp.arange(6)
      if indices_are_sorted:
        segment_ids = jnp.array([0, 1, 2, 3, 4, 5])
        expected_out = jnp.array([0, 1, 2, 3, 4, 5])
        num_segments = 6
      else:
        segment_ids = jnp.array([1, 0, 2, 4, 3, -5])
        expected_out = jnp.array([1, 0, 2, 4, 3])
        num_segments = 5
    else:
      data = jnp.arange(9)
      if indices_are_sorted:
        segment_ids = jnp.array([0, 0, 0, 1, 1, 1, 2, 3, 4])
        expected_out = jnp.array([0, 3, 6, 7, 8, inf])
      else:
        segment_ids = jnp.array([0, 1, 2, 0, 4, 0, 1, 1, -6])
        expected_out = jnp.array([0, 1, 2, inf, 4, inf])
      num_segments = 6

    with self.subTest('nojit'):
      result = utils.segment_min(data, segment_ids, num_segments,
                                 indices_are_sorted, unique_indices)
      np.testing.assert_allclose(result, expected_out)
      result = utils.segment_min(data, segment_ids,
                                 indices_are_sorted=indices_are_sorted,
                                 unique_indices=unique_indices)
      num_unique_segment = np.max(segment_ids) + 1
      np.testing.assert_allclose(result, expected_out[:num_unique_segment])
    with self.subTest('jit'):
      result = jax.jit(utils.segment_min, static_argnums=(2, 3, 4))(
          data, segment_ids, num_segments, indices_are_sorted, unique_indices)
      np.testing.assert_allclose(result, expected_out)

  @parameterized.parameters((False, False), (True, False), (True, True),
                            (False, True))
  def test_segment_min_or_constant(self, indices_are_sorted, unique_indices):
    if unique_indices:
      data = jnp.arange(6, dtype=jnp.float32)
      if indices_are_sorted:
        segment_ids = jnp.array([0, 1, 2, 3, 4, 5])
        expected_out = jnp.array([0, 1, 2, 3, 4, 5], dtype=jnp.float32)
        num_segments = 6
      else:
        segment_ids = jnp.array([1, 0, 2, 4, 3, -5])
        expected_out = jnp.array([1, 0, 2, 4, 3], dtype=jnp.float32)
        num_segments = 5
    else:
      data = jnp.arange(9, dtype=jnp.float32)
      if indices_are_sorted:
        segment_ids = jnp.array([0, 0, 0, 1, 1, 1, 2, 3, 4])
        expected_out = jnp.array([0, 3, 6, 7, 8, 0], dtype=jnp.float32)
      else:
        segment_ids = jnp.array([0, 1, 2, 0, 4, 0, 1, 1, -6])
        expected_out = jnp.array([0, 1, 2, 0, 4, 0], dtype=jnp.float32)
      num_segments = 6

    with self.subTest('nojit'):
      result = utils.segment_min_or_constant(data, segment_ids, num_segments,
                                             indices_are_sorted, unique_indices)
      np.testing.assert_allclose(result, expected_out)
      grad = jax.grad(lambda *x: jnp.sum(utils.segment_min_or_constant(*x)))(
          data, segment_ids, num_segments, indices_are_sorted, unique_indices)
      assert np.all(jnp.isfinite(grad))
      result = utils.segment_min_or_constant(
          data,
          segment_ids,
          indices_are_sorted=indices_are_sorted,
          unique_indices=unique_indices)
      num_unique_segments = jnp.max(segment_ids) + 1
      np.testing.assert_allclose(result, expected_out[:num_unique_segments])
    with self.subTest('jit'):
      result = jax.jit(
          utils.segment_min_or_constant,
          static_argnums=(2, 3, 4))(data, segment_ids, num_segments,
                                    indices_are_sorted, unique_indices)
      np.testing.assert_allclose(result, expected_out)
      grad_fn = jax.jit(
          jax.grad(lambda *x: jnp.sum(utils.segment_min_or_constant(*x))),
          static_argnums=(2, 3, 4))
      grad = grad_fn(data, segment_ids, num_segments, indices_are_sorted,
                     unique_indices)
      assert np.all(jnp.isfinite(grad))

  @parameterized.parameters((False, False), (True, False), (True, True),
                            (False, True))
  def test_segment_min_or_constant_2d(self, indices_are_sorted, unique_indices):
    if unique_indices:
      data = jnp.stack([jnp.arange(6), jnp.arange(6, 0, -1)], axis=1)
      if indices_are_sorted:
        segment_ids = jnp.array([0, 1, 2, 3, 4, 5])
        expected_out = jnp.array(
            [[0, 6], [1, 5], [2, 4], [3, 3], [4, 2], [5, 1]])
        num_segments = 6
      else:
        segment_ids = jnp.array([1, 0, 2, 4, 3, -5])
        expected_out = jnp.array(
            [[1, 5], [0, 6], [2, 4], [4, 2], [3, 3]])
        num_segments = 5
    else:
      data = jnp.stack([jnp.arange(9), jnp.arange(9, 0, -1)], axis=1)
      if indices_are_sorted:
        segment_ids = jnp.array([0, 0, 0, 1, 1, 1, 2, 3, 4])
        expected_out = jnp.array(
            [[0, 7], [3, 4], [6, 3], [7, 2], [8, 1], [0, 0]])
      else:
        segment_ids = jnp.array([0, 1, 2, 0, 4, 0, 1, 1, -6])
        expected_out = jnp.array(
            [[0, 4], [1, 2], [2, 7], [0, 0], [4, 5], [0, 0]])
      num_segments = 6

    with self.subTest('nojit'):
      result = utils.segment_min_or_constant(data, segment_ids, num_segments,
                                             indices_are_sorted, unique_indices)
      np.testing.assert_allclose(result, expected_out)
      result = utils.segment_min_or_constant(
          data,
          segment_ids,
          indices_are_sorted=indices_are_sorted,
          unique_indices=unique_indices)
      num_unique_segments = jnp.max(segment_ids) + 1
      np.testing.assert_allclose(result, expected_out[:num_unique_segments])
    with self.subTest('jit'):
      result = jax.jit(utils.segment_min_or_constant, static_argnums=(2, 3, 4))(
          data, segment_ids, num_segments, indices_are_sorted, unique_indices)
      np.testing.assert_allclose(result, expected_out)

  @parameterized.parameters((True,), (False,))
  def test_segment_softmax(self, nan_data):
    data = jnp.arange(9, dtype=jnp.float32)
    segment_ids = jnp.array([0, 1, 2, 0, 4, 0, 1, 1, 0])
    num_segments = 6
    expected_out = jnp.array([3.1741429e-04, 1.8088353e-03, 1.0000000e+00,
                              6.3754367e-03, 1.0000000e+00, 4.7108460e-02,
                              2.6845494e-01, 7.2973621e-01, 9.4619870e-01])
    if nan_data:
      data = data.at[0].set(jnp.nan)
      expected_out = expected_out.at[np.array([0, 3, 5, 8])].set(jnp.nan)
    with self.subTest('nojit'):
      result = utils.segment_softmax(data, segment_ids, num_segments)
      np.testing.assert_allclose(result, expected_out)
      result = utils.segment_softmax(data, segment_ids)
      np.testing.assert_allclose(result, expected_out)
    with self.subTest('jit'):
      result = jax.jit(
          utils.segment_softmax, static_argnums=2)(data, segment_ids,
                                                   num_segments)
      np.testing.assert_allclose(result, expected_out)

  def test_partition_softmax(self):
    data = jnp.arange(9)
    partitions = jnp.array([3, 2, 4])
    sum_partitions = 9
    expected_out = np.array([0.090031, 0.244728, 0.665241, 0.268941, 0.731059,
                             0.032059, 0.087144, 0.236883, 0.643914])
    with self.subTest('nojit'):
      result = utils.partition_softmax(data, partitions, sum_partitions)
      jax.tree_util.tree_map(
          functools.partial(np.testing.assert_allclose, atol=1E-5, rtol=1E-5),
          result, expected_out)
      result = utils.partition_softmax(data, partitions)
      jax.tree_util.tree_map(
          functools.partial(np.testing.assert_allclose, atol=1E-5, rtol=1E-5),
          result, expected_out)
    with self.subTest('jit'):
      result = jax.jit(utils.partition_softmax, static_argnums=2)(
          data, partitions, sum_partitions)
      jax.tree_util.tree_map(
          functools.partial(np.testing.assert_allclose, atol=1E-5, rtol=1E-5),
          result, expected_out)

  @parameterized.named_parameters(('valid_1_no_feat', 1, 1, False, False),
                                  ('valid_5_no_feat', 5, 5, False, False),
                                  ('valid_1_nodes', 1, 1, True, False),
                                  ('valid_5_globals', 5, 5, False, True),
                                  ('valid_5_both', 5, 5, True, True),
                                  ('zero_nodes', 0, 1, False, False),
                                  ('zero_graphs', 1, 0, False, False),)
  def test_fully_connected_graph(self, n_node, n_graph, nodes, globals_):
    node_feat = np.random.rand(n_node*n_graph, 32) if nodes else None
    global_feat = np.random.rand(n_graph, 32) if globals_ else None
    with self.subTest('nojit'):
      result = utils.get_fully_connected_graph(
          n_node, n_graph, node_feat, global_feat)
      if nodes:
        self.assertLen(result.nodes, n_node*n_graph)
      if globals_:
        self.assertLen(result.globals, n_graph)
      self.assertLen(result.senders, n_node**2 * n_graph)
      self.assertLen(result.receivers, n_node**2 * n_graph)
      np.testing.assert_allclose(result.n_node, jnp.array([n_node] * n_graph))
    with self.subTest('jit'):
      result = jax.jit(utils.get_fully_connected_graph, static_argnums=[0, 1])(
          n_node, n_graph, node_feat, global_feat)
      if nodes:
        self.assertLen(result.nodes, n_node*n_graph)
      if globals_:
        self.assertLen(result.globals, n_graph)
      self.assertLen(result.senders, n_node**2 * n_graph)
      self.assertLen(result.receivers, n_node**2 * n_graph)
      np.testing.assert_allclose(result.n_node, jnp.array([n_node] * n_graph))

    with self.subTest('senders_receiver_indices'):
      if n_node > 0:
        # [0, 1, ..., n_node - 1]
        node_indices = jnp.arange(n_node)
        # [0, 1,..., n_node - 1] + [0, 1,..., n_node - 1] + ... n_node times
        # [0,..., 0, 1,..., 1,..., n_node - 1,..., n_node - 1] each n_node times
        expected_senders = np.concatenate([node_indices] * n_node, axis=0)
        expected_receivers = np.stack(
            [node_indices] * n_node, axis=-1).reshape([-1])
      else:
        expected_senders = np.array([], dtype=np.int32)
        expected_receivers = np.array([], dtype=np.int32)

      # Check sender and receivers on each graph in the batch.
      for result_graph in utils.unbatch(result):
        np.testing.assert_allclose(result_graph.senders, expected_senders)
        np.testing.assert_allclose(result_graph.receivers, expected_receivers)

  @parameterized.named_parameters(('valid_1_no_feat', 1, 1),
                                  ('valid_5_no_feat', 5, 5),
                                  ('zero_nodes', 0, 1),
                                  ('zero_graphs', 1, 0),)
  def test_fully_connected_graph_no_self_edges(self, n_node, n_graph):

    # `test_fully_connected_graph` already tests the case `add_self_edges=True`
    # so all that is left to test is that if we set `add_self_edges=False` we
    # get the same edges, except for the self-edges (although order may differ).
    graph_with_self_edges = utils.get_fully_connected_graph(
        n_node, n_graph, add_self_edges=True)
    graph_without_self_edges = utils.get_fully_connected_graph(
        n_node, n_graph, add_self_edges=False)

    # We will use sets to compare the order, since the order is not preserved
    # due to the usage of `np.roll` (e.g. if you remove the self edges after
    # add_self_edges=True, the remaining edges are in a different order than if
    # add_self_edges=False).
    send_recv_actual = {
        (s, r) for s, r in zip(
            np.asarray(graph_without_self_edges.senders),
            np.asarray(graph_without_self_edges.receivers))}

    # Remove the self edges by hand from `graph_with_self_edges`
    mask_self_edges = (
        graph_with_self_edges.senders == graph_with_self_edges.receivers)
    send_recv_expected = {
        (s, r) for s, r in zip(
            np.asarray(graph_with_self_edges.senders[~mask_self_edges]),
            np.asarray(graph_with_self_edges.receivers[~mask_self_edges]))}
    self.assertSetEqual(send_recv_actual, send_recv_expected)

  @parameterized.named_parameters(('with_self_edges', True),
                                  ('without_self_edges', False),)
  def test_fully_connected_graph_order_edges(self, add_self_edges):
    # This helps documenting the order of the output edges, so we are aware
    # in case we accidentally change it.
    graph_batch = utils.get_fully_connected_graph(
        n_node_per_graph=3,
        n_graph=1,
        add_self_edges=add_self_edges)

    if add_self_edges:
      np.testing.assert_array_equal(
          graph_batch.senders, [0, 1, 2] * 3)
      np.testing.assert_array_equal(
          graph_batch.receivers, [0] * 3 + [1] * 3 + [2] * 3)
    else:
      np.testing.assert_array_equal(graph_batch.senders, [1, 2, 2, 0, 0, 1])
      np.testing.assert_array_equal(graph_batch.receivers, [0, 0, 1, 1, 2, 2])


class ConcatenatedArgsWrapperTest(parameterized.TestCase):

  @parameterized.parameters(
      ([], {'a': np.array([10, 2])}, -1),
      ([np.array([10, 5])], {'a': np.array([10, 2])}, -1),
      ([np.array([10, 5]), np.array([10, 3])], {'a': np.array([10, 2])}, -1),
      ([np.array([10, 5]), np.array([10, 3])], {}, -1),
      ([{'a': np.array([10, 2]), 'b': np.array([10, 4])}],
       {'c': np.array([10, 3])}, 1),
      ([{'a': np.array([2, 10]), 'b': np.array([4, 10])}],
       {'c': np.array([3, 10])}, 0))
  def test_single_arg(self, args_shapes, kwargs_shapes, axis):
    args = tree.tree_map(lambda x: np.random.randn(*x), args_shapes)
    kwargs = {k: np.random.randn(*shape) for k, shape in kwargs_shapes.items()}

    @utils.concatenated_args(axis=axis)
    def update_fn(feat):
      return feat

    out = update_fn(*args, **kwargs)
    expected_out = jnp.concatenate(
        list(tree.tree_flatten(args)[0]) + list(tree.tree_flatten(kwargs)[0]),
        axis=axis)
    np.testing.assert_allclose(out, expected_out)


_DB_NUM_NODES = (10, 15)
_DB_NODE_SHAPE = (3, 4, 1)
_DB_NUM_EDGES = (12, 17)
_DB_EDGE_SHAPE = (4, 3)
_DB_GLOBAL_SHAPE = (2, 3, 1, 4)


def _make_dynamic_batch_graph(
    add_globals,
    num_nodes=_DB_NUM_NODES,
    num_edges=_DB_NUM_EDGES,
):
  total_num_edges = sum(num_edges)
  total_num_nodes = sum(num_nodes)
  g_ = _make_nest(
      np.random.normal(size=_DB_GLOBAL_SHAPE)) if add_globals else {}
  return graph.GraphsTuple(
      nodes=_make_nest(
          np.random.normal(size=(total_num_nodes,) + _DB_NODE_SHAPE)),
      edges=_make_nest(
          np.random.normal(size=(total_num_edges,) + _DB_EDGE_SHAPE)),
      n_edge=np.array(num_edges),
      n_node=np.array(num_nodes),
      senders=np.random.randint(
          0, total_num_nodes, size=total_num_edges, dtype=np.int32),
      receivers=np.random.randint(
          0, total_num_nodes, size=total_num_edges, dtype=np.int32),
      globals=g_)


class DynamicBatchTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    os.environ['XLA_FLAGS'] = '--xla_force_host_platform_device_count=4'
    xla_bridge.get_backend.cache_clear()
    self._global_graph = _make_dynamic_batch_graph(add_globals=True)
    self._global_small_graph = _make_dynamic_batch_graph(
        add_globals=True, num_nodes=(5, 7), num_edges=(6, 8))

  @parameterized.named_parameters(
      ('graph_with_globals_n_node_hit', True, {
          'n_node': sum(_DB_NUM_NODES) + 1,
          'n_edge': sum(_DB_NUM_EDGES) + 100,
          'n_graph': len(_DB_NUM_EDGES) + 100
      }),
      ('graph_with_globals_n_edge_hit', True, {
          'n_node': sum(_DB_NUM_NODES) + 100,
          'n_edge': sum(_DB_NUM_EDGES),
          'n_graph': len(_DB_NUM_EDGES) + 100
      }),
      ('graph_with_globals_n_graph_hit', True, {
          'n_node': sum(_DB_NUM_NODES) + 100,
          'n_edge': sum(_DB_NUM_EDGES) + 100,
          'n_graph': len(_DB_NUM_EDGES) + 1
      }),
      (
          'graph_with_globals_no_budget_hit',
          False,
          {
              # Add enough padding so not enough for a single extra graph.
              'n_node': sum(_DB_NUM_NODES) + 5,
              'n_edge': sum(_DB_NUM_EDGES) + 5,
              'n_graph': len(_DB_NUM_EDGES) + 5
          }),
      (
          'graph_no_globals_no_budget_hit',
          False,
          {
              # Add enough padding so not enough for a single extra graph.
              'n_node': sum(_DB_NUM_NODES) + 5,
              'n_edge': sum(_DB_NUM_EDGES) + 5,
              'n_graph': len(_DB_NUM_EDGES) + 5
          }),
      (
          'graph_nests_no_globals_no_budget_hit',
          False,
          {
              # Add enough padding so not enough for a single extra graph.
              'n_node': sum(_DB_NUM_NODES) + 5,
              'n_edge': sum(_DB_NUM_EDGES) + 5,
              'n_graph': len(_DB_NUM_EDGES) + 5
          }))
  def test_dynamically_batch(self, use_globals, batch_kwargs):

    def graph_iterator():
      graphs = [
          _make_dynamic_batch_graph(add_globals=use_globals) for x in range(4)]
      return iter(graphs + utils.unbatch_np(graphs[-1]))

    batched_dataset = utils.dynamically_batch(graph_iterator(),
                                              **batch_kwargs)
    batched_graphs = []
    while True:
      try:
        batched_graphs.append(next(batched_dataset))
      except StopIteration:
        break

    self.assertLen(batched_graphs, 5)
    for batch_graphs in batched_graphs:
      batch_nodes = jax.tree_util.tree_flatten(batch_graphs.nodes)[0]
      for nodes in batch_nodes:
        self.assertEqual(nodes.shape[0], batch_kwargs['n_node'])
      batch_edges = jax.tree_util.tree_flatten(batch_graphs.edges)[0]
      for edges in batch_edges:
        self.assertEqual(edges.shape[0], batch_kwargs['n_edge'])
      self.assertLen(batch_graphs.n_node, batch_kwargs['n_graph'])
      self.assertEqual(
          utils.get_number_of_padding_with_graphs_nodes(batch_graphs),
          batch_kwargs['n_node'] - sum(_DB_NUM_NODES))
      self.assertEqual(
          utils.get_number_of_padding_with_graphs_edges(batch_graphs),
          batch_kwargs['n_edge'] - sum(_DB_NUM_EDGES))

  def test_too_big_graphs_tuple(self):
    with self.subTest('test_too_big_nodes'):
      iterator = utils.dynamically_batch(
          iter([self._global_graph]), n_node=15, n_edge=50, n_graph=10)
      self.assertRaisesRegex(
          RuntimeError, 'Found graph bigger than batch size.*',
          lambda: next(iterator))
    with self.subTest('test_too_big_edges'):
      iterator = utils.dynamically_batch(
          iter([self._global_graph]), n_node=26, n_edge=15, n_graph=10)
      self.assertRaisesRegex(
          RuntimeError, 'Found graph bigger than batch size.*',
          lambda: next(iterator))
    with self.subTest('test_too_big_graphs'):
      iterator = utils.dynamically_batch(
          iter([self._global_graph]), n_node=50, n_edge=50, n_graph=1)
      self.assertRaisesRegex(
          ValueError, 'The number of graphs*',
          lambda: next(iterator))
    with self.subTest('test_too_big_fails_gracefully'):
      # Ensure that dynamically_batch() returns the accumulated batch before
      # raising an exception.
      iterator = utils.dynamically_batch(
          iter([self._global_small_graph, self._global_graph]),
          n_node=15, n_edge=15, n_graph=10)
      next(iterator)
      self.assertRaisesRegex(
          RuntimeError, 'Found graph bigger than batch size.*',
          lambda: next(iterator))

  def test_not_enough_graphs(self):
    iterator = utils.dynamically_batch(
        iter([self._global_graph]), n_node=5, n_edge=5, n_graph=1)
    self.assertRaisesRegex(
        ValueError, 'The number of graphs*', lambda: next(iterator))


class ZeroOutTest(parameterized.TestCase):

  def _assert_values_for_graph(self, padded_graph, wrapper):
    # Make padded graph values non zero.
    padded_graph = padded_graph._replace(
        nodes=tree.tree_map(lambda x: x - 1., padded_graph.nodes),
        edges=tree.tree_map(lambda x: x - 1., padded_graph.edges),
        globals=tree.tree_map(lambda x: x - 1., padded_graph.globals))
    true_valid_graph = utils.unbatch(padded_graph)[0]
    if wrapper:
      zeroed_graph_net = utils.with_zero_out_padding_outputs(lambda x: x)
      zeroed_padded_graph = zeroed_graph_net(padded_graph)
    else:
      zeroed_padded_graph = utils.zero_out_padding(padded_graph)
    graphs = utils.unbatch(zeroed_padded_graph)
    valid_graph = graphs[0]
    padding_graphs = graphs[1:]
    tree.tree_map(np.testing.assert_array_equal, valid_graph.nodes,
                  true_valid_graph.nodes)
    for padding_graph in padding_graphs:
      tree.tree_map(
          lambda x: np.testing.assert_array_equal(x, jnp.zeros_like(x)),
          padding_graph.nodes)

  @parameterized.parameters(True, False)
  def test_zero_padding_values(self, wrapper):
    g = _get_random_graph(max_n_graph=1)
    with self.subTest('test_all_padded_features'):
      self._assert_values_for_graph(
          utils.pad_with_graphs(g, n_node=20, n_edge=20, n_graph=3),
          wrapper=wrapper)
    with self.subTest('test_no_edge_features'):
      self._assert_values_for_graph(
          utils.pad_with_graphs(
              g, n_node=sum(g.n_node) + 1, n_edge=sum(g.n_edge), n_graph=3),
          wrapper=wrapper)
    with self.subTest('test_no_extra_graph_features'):
      self._assert_values_for_graph(
          utils.pad_with_graphs(
              g, n_node=sum(g.n_node) + 1, n_edge=sum(g.n_edge), n_graph=2),
          wrapper=wrapper)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for jraph.models."""

from absl.testing import absltest
from absl.testing import parameterized
import jax
import jax.numpy as jnp
import jax.tree_util as tree

from jraph._src import graph
from jraph._src import models
from jraph._src import utils
import numpy as np


def _get_random_graph(max_n_graph=10):
  n_graph = np.random.randint(1, max_n_graph + 1)
  n_node = np.random.randint(0, 10, n_graph)
  n_edge = np.random.randint(0, 20, n_graph)
  # We cannot have any edges if there are no nodes.
  n_edge[n_node == 0] = 0

  senders = []
  receivers = []
  offset = 0
  for n_node_in_graph, n_edge_in_graph in zip(n_node, n_edge):
    if n_edge_in_graph != 0:
      senders += list(
          np.random.randint(0, n_node_in_graph, n_edge_in_graph) + offset)
      receivers += list(
          np.random.randint(0, n_node_in_graph, n_edge_in_graph) + offset)
    offset += n_node_in_graph

  return graph.GraphsTuple(
      n_node=jnp.asarray(n_node),
      n_edge=jnp.asarray(n_edge),
      nodes=jnp.asarray(np.random.random(size=(np.sum(n_node), 4))),
      edges=jnp.asarray(np.random.random(size=(np.sum(n_edge), 3))),
      globals=jnp.asarray(np.random.random(size=(n_graph, 5))),
      senders=jnp.asarray(senders),
      receivers=jnp.asarray(receivers))


def _get_graph_network(graphs_tuple):
  # Our test update functions are just identity functions.
  update_node_fn = lambda n, se, re, g: n
  update_edge_fn = lambda e, sn, rn, g: e
  update_global_fn = lambda gn, ge, g: g
  net = models.GraphNetwork(update_edge_fn,
                            update_node_fn,
                            update_global_fn)
  return net(graphs_tuple)


def _get_graph_network_no_global_update(graphs_tuple):
  # Our test update functions are just identity functions.
  update_node_fn = lambda n, se, re, g: n
  update_edge_fn = lambda e, sn, rn, g: e
  update_global_fn = None
  net = models.GraphNetwork(update_edge_fn,
                            update_node_fn,
                            update_global_fn)
  return net(graphs_tuple)


def _get_graph_network_no_node_update(graphs_tuple):
  # Our test update functions are just identity functions.
  update_node_fn = None
  update_edge_fn = lambda e, sn, rn, g: e
  update_global_fn = lambda gn, ge, g: g
  net = models.GraphNetwork(update_edge_fn,
                            update_node_fn,
                            update_global_fn)
  return net(graphs_tuple)


def _get_graph_network_no_edge_update(graphs_tuple):
  # Our test update functions are just identity functions.
  update_node_fn = lambda n, se, re, g: n
  update_edge_fn = None
  update_global_fn = lambda gn, ge, g: g
  net = models.GraphNetwork(update_edge_fn,
                            update_node_fn,
                            update_global_fn)
  return net(graphs_tuple)


def _get_attention_graph_network(graphs_tuple):
  # Our test update functions are just identity functions.
  update_node_fn = lambda n, se, re, g: n
  update_edge_fn = lambda e, sn, rn, g: e
  update_global_fn = lambda gn, ge, g: g
  # Our attention logits are just one in this case.
  attention_logit_fn = lambda e, sn, rn, g: jnp.array(1.0)
  # We use a custom apply function here, which just returns the edge unchanged.
  attention_reduce_fn = lambda e, w: e
  net = models.GraphNetwork(update_edge_fn,
                            update_node_fn,
                            update_global_fn,
                            attention_logit_fn=attention_logit_fn,
                            attention_reduce_fn=attention_reduce_fn)
  return net(graphs_tuple)


def _get_graph_gat(graphs_tuple):
  # Our test update functions are just identity functions.
  update_node_fn = lambda n, se, re, g: n
  update_edge_fn = lambda e, sn, rn, g: e
  update_global_fn = lambda gn, ge, g: g
  # Our attention logits are just one in this case.
  attention_logit_fn = lambda e, sn, rn, g: jnp.array(1.0)
  # We use a custom apply function here, which just returns the edge unchanged.
  attention_reduce_fn = lambda e, w: e
  net = models.GraphNetGAT(update_edge_fn,
                           update_node_fn,
                           attention_logit_fn,
                           attention_reduce_fn,
                           update_global_fn)
  return net(graphs_tuple)


def _get_multi_head_attention_graph_network(graphs_tuple):
  # Our test update functions are just identity functions.
  update_node_fn = lambda n, se, re, g: n
  update_global_fn = lambda gn, ge, g: g
  # With multi-head attention we have to return multiple edge features.
  # Here we define 3 heads, all with the same message.
  def update_edge_fn(e, unused_sn, unused_rn, unused_g):
    return tree.tree_map(lambda e_: jnp.stack([e_, e_, e_]), e)
  # Our attention logits are just the sum of the edge features of each head.
  def attention_logit_fn(e, unused_sn, unused_rn, unused_g):
    return tree.tree_map(lambda e_: jnp.sum(e_, axis=-1), e)
  # For multi-head attention we need a custom apply attention function.
  # In this we return the first edge.
  def attention_reduce_fn(e, unused_w):
    return tree.tree_map(lambda e_: e_[0], e)
  net = models.GraphNetwork(jax.vmap(update_edge_fn),
                            jax.vmap(update_node_fn),
                            update_global_fn,
                            attention_logit_fn=jax.vmap(attention_logit_fn),
                            attention_reduce_fn=jax.vmap(attention_reduce_fn))
  return net(graphs_tuple)


def _get_interaction_network(graphs_tuple):
  update_node_fn = lambda n, r: jnp.concatenate((n, r), axis=-1)
  update_edge_fn = lambda e, s, r: jnp.concatenate((e, s, r), axis=-1)
  out = models.InteractionNetwork(update_edge_fn, update_node_fn)(graphs_tuple)
  nodes, edges, receivers, senders, _, _, _ = graphs_tuple
  expected_edges = jnp.concatenate(
      (edges, nodes[senders], nodes[receivers]), axis=-1)
  aggregated_nodes = utils.segment_sum(
      expected_edges, receivers, num_segments=len(graphs_tuple.nodes))
  expected_nodes = jnp.concatenate(
      (nodes, aggregated_nodes), axis=-1)
  expected_out = graphs_tuple._replace(
      edges=expected_edges, nodes=expected_nodes)
  return out, expected_out


def _get_graph_independent(graphs_tuple):
  embed_fn = lambda x: x * 2
  out = models.GraphMapFeatures(embed_fn, embed_fn, embed_fn)(graphs_tuple)
  expected_out = graphs_tuple._replace(nodes=graphs_tuple.nodes*2,
                                       edges=graphs_tuple.edges*2,
                                       globals=graphs_tuple.globals*2)
  return out, expected_out


def _get_relation_network(graphs_tuple):
  edge_fn = lambda s, r: jnp.concatenate((s, r), axis=-1)
  global_fn = lambda e: e*2
  out = models.RelationNetwork(edge_fn, global_fn)(graphs_tuple)
  expected_edges = jnp.concatenate(
      (graphs_tuple.nodes[graphs_tuple.senders],
       graphs_tuple.nodes[graphs_tuple.receivers]), axis=-1)
  num_graphs = len(graphs_tuple.n_edge)
  edge_gr_idx = jnp.repeat(jnp.arange(num_graphs),
                           graphs_tuple.n_edge,
                           total_repeat_length=graphs_tuple.edges.shape[0])
  aggregated_edges = utils.segment_sum(
      expected_edges, edge_gr_idx, num_segments=num_graphs)
  expected_out = graphs_tuple._replace(
      edges=expected_edges, globals=aggregated_edges*2)
  return out, expected_out


def _get_deep_sets(graphs_tuple):
  node_fn = lambda n, g: jnp.concatenate((n, g), axis=-1)
  global_fn = lambda n: n*2
  out = models.DeepSets(node_fn, global_fn)(graphs_tuple)
  num_graphs = len(graphs_tuple.n_node)
  num_nodes = len(graphs_tuple.nodes)
  broadcasted_globals = jnp.repeat(graphs_tuple.globals, graphs_tuple.n_node,
                                   total_repeat_length=num_nodes, axis=0)
  expected_nodes = jnp.concatenate(
      (graphs_tuple.nodes, broadcasted_globals), axis=-1)
  node_gr_idx = jnp.repeat(jnp.arange(num_graphs),
                           graphs_tuple.n_node,
                           total_repeat_length=num_nodes)
  expected_out = graphs_tuple._replace(
      nodes=expected_nodes,
      globals=utils.segment_sum(
          expected_nodes, node_gr_idx, num_segments=num_graphs)*2)
  return out, expected_out


def _get_gat(graphs_tuple):
  # With multi-head attention we have to return multiple edge features.
  # Here we define 3 heads, all with the same message.
  def attention_query_fn(n):
    return tree.tree_map(lambda n_: jnp.stack([n_, n_, n_], axis=2), n)
  # Our attention logits 1 if a self edge
  def attention_logit_fn(s, r, e_):
    del e_
    return (s == r)*1 + (s != r)*-1e10

  def node_update_fn(nodes):
    return jnp.mean(nodes, axis=2)

  net = models.GAT(attention_query_fn, attention_logit_fn, node_update_fn)

  # Cast nodes to floats since GAT will output floats from the softmax
  # attention.
  graphs_tuple = graphs_tuple._replace(
      nodes=jnp.array(graphs_tuple.nodes, jnp.float32))
  return net(graphs_tuple), graphs_tuple


class ModelsTest(parameterized.TestCase):

  def _make_nest(self, array):
    """Returns a nest given an array."""
    return {'a': array,
            'b': [jnp.ones_like(array), {'c': jnp.zeros_like(array)}]}

  def _get_list_and_batched_graph(self):
    """Returns a list of individual graphs and a batched version.

    This test-case includes the following corner-cases:
      - single node,
      - multiple nodes,
      - no edges,
      - single edge,
      - and multiple edges.
    """
    batched_graph = graph.GraphsTuple(
        n_node=jnp.array([1, 3, 1, 0, 2, 0, 0]),
        n_edge=jnp.array([2, 5, 0, 0, 1, 0, 0]),
        nodes=self._make_nest(jnp.arange(14).reshape(7, 2)),
        edges=self._make_nest(jnp.arange(24).reshape(8, 3)),
        globals=self._make_nest(jnp.arange(14).reshape(7, 2)),
        senders=jnp.array([0, 0, 1, 1, 2, 3, 3, 6]),
        receivers=jnp.array([0, 0, 2, 1, 3, 2, 1, 5]))

    list_graphs = [
        graph.GraphsTuple(
            n_node=jnp.array([1]),
            n_edge=jnp.array([2]),
            nodes=self._make_nest(jnp.array([[0, 1]])),
            edges=self._make_nest(jnp.array([[0, 1, 2], [3, 4, 5]])),
            globals=self._make_nest(jnp.array([[0, 1]])),
            senders=jnp.array([0, 0]),
            receivers=jnp.array([0, 0])),
        graph.GraphsTuple(
            n_node=jnp.array([3]),
            n_edge=jnp.array([5]),
            nodes=self._make_nest(jnp.array([[2, 3], [4, 5], [6, 7]])),
            edges=self._make_nest(
                jnp.array([[6, 7, 8], [9, 10, 11], [12, 13, 14], [15, 16, 17],
                           [18, 19, 20]])),
            globals=self._make_nest(jnp.array([[2, 3]])),
            senders=jnp.array([0, 0, 1, 2, 2]),
            receivers=jnp.array([1, 0, 2, 1, 0])),
        graph.GraphsTuple(
            n_node=jnp.array([1]),
            n_edge=jnp.array([0]),
            nodes=self._make_nest(jnp.array([[8, 9]])),
            edges=self._make_nest(jnp.zeros((0, 3))),
            globals=self._make_nest(jnp.array([[4, 5]])),
            senders=jnp.array([]),
            receivers=jnp.array([])),
        graph.GraphsTuple(
            n_node=jnp.array([0]),
            n_edge=jnp.array([0]),
            nodes=self._make_nest(jnp.zeros((0, 2))),
            edges=self._make_nest(jnp.zeros((0, 3))),
            globals=self._make_nest(jnp.array([[6, 7]])),
            senders=jnp.array([]),
            receivers=jnp.array([])),
        graph.GraphsTuple(
            n_node=jnp.array([2]),
            n_edge=jnp.array([1]),
            nodes=self._make_nest(jnp.array([[10, 11], [12, 13]])),
            edges=self._make_nest(jnp.array([[21, 22, 23]])),
            globals=self._make_nest(jnp.array([[8, 9]])),
            senders=jnp.array([1]),
            receivers=jnp.array([0])),
        graph.GraphsTuple(
            n_node=jnp.array([0]),
            n_edge=jnp.array([0]),
            nodes=self._make_nest(jnp.zeros((0, 2))),
            edges=self._make_nest(jnp.zeros((0, 3))),
            globals=self._make_nest(jnp.array([[10, 11]])),
            senders=jnp.array([]),
            receivers=jnp.array([])),
        graph.GraphsTuple(
            n_node=jnp.array([0]),
            n_edge=jnp.array([0]),
            nodes=self._make_nest(jnp.zeros((0, 2))),
            edges=self._make_nest(jnp.zeros((0, 3))),
            globals=self._make_nest(jnp.array([[12, 13]])),
            senders=jnp.array([]),
            receivers=jnp.array([]))
    ]

    return list_graphs, batched_graph

  @parameterized.parameters(_get_graph_network,
                            _get_graph_network_no_node_update,
                            _get_graph_network_no_edge_update,
                            _get_graph_network_no_global_update,
                            _get_attention_graph_network,
                            _get_multi_head_attention_graph_network,
                            _get_graph_gat)
  def test_connect_graphnetwork(self, network_fn):
    _, batched_graphs_tuple = self._get_list_and_batched_graph()
    with self.subTest('nojit'):
      out = network_fn(batched_graphs_tuple)
      jax.tree_util.tree_map(np.testing.assert_allclose, out,
                             batched_graphs_tuple)
    with self.subTest('jit'):
      out = jax.jit(network_fn)(batched_graphs_tuple)
      jax.tree_util.tree_map(np.testing.assert_allclose, out,
                             batched_graphs_tuple)

  @parameterized.parameters(_get_graph_network,
                            _get_graph_network_no_node_update,
                            _get_graph_network_no_edge_update,
                            _get_graph_network_no_global_update)
  def test_connect_graphnetwork_nones(self, network_fn):
    batched_graphs_tuple = graph.GraphsTuple(
        n_node=jnp.array([1, 3, 1, 0, 2, 0, 0]),
        n_edge=jnp.array([2, 5, 0, 0, 1, 0, 0]),
        nodes=self._make_nest(jnp.arange(14).reshape(7, 2)),
        edges=self._make_nest(jnp.arange(24).reshape(8, 3)),
        globals=self._make_nest(jnp.arange(14).reshape(7, 2)),
        senders=jnp.array([0, 0, 1, 1, 2, 3, 3, 6]),
        receivers=jnp.array([0, 0, 2, 1, 3, 2, 1, 5]))

    for name, graphs_tuple in [
        ('no_globals', batched_graphs_tuple._replace(globals=None)),
        ('empty_globals', batched_graphs_tuple._replace(globals=[])),
        ('no_edges', batched_graphs_tuple._replace(edges=None)),
        ('empty_edges', batched_graphs_tuple._replace(edges=[])),
    ]:
      with self.subTest(name + '_nojit'):
        out = network_fn(graphs_tuple)
        jax.tree_util.tree_map(np.testing.assert_allclose, out, graphs_tuple)
      with self.subTest(name + '_jit'):
        out = jax.jit(network_fn)(graphs_tuple)
        jax.tree_util.tree_map(np.testing.assert_allclose, out, graphs_tuple)

  @parameterized.parameters(_get_interaction_network,
                            _get_graph_independent,
                            _get_gat,
                            _get_relation_network,
                            _get_deep_sets)
  def test_connect_gnns(self, network_fn):
    batched_graphs_tuple = graph.GraphsTuple(
        n_node=jnp.array([1, 3, 1, 0, 2, 0, 0]),
        n_edge=jnp.array([1, 7, 1, 0, 3, 0, 0]),
        nodes=jnp.arange(14).reshape(7, 2),
        edges=jnp.arange(36).reshape(12, 3),
        globals=jnp.arange(14).reshape(7, 2),
        senders=jnp.array([0, 1, 2, 3, 4, 5, 6, 1, 2, 3, 3, 6]),
        receivers=jnp.array([0, 1, 2, 3, 4, 5, 6, 2, 3, 2, 1, 5]))
    with self.subTest('nojit'):
      out, expected_out = network_fn(batched_graphs_tuple)
      jax.tree_util.tree_map(np.testing.assert_allclose, out, expected_out)
    with self.subTest('jit'):
      out, expected_out = jax.jit(network_fn)(batched_graphs_tuple)
      jax.tree_util.tree_map(np.testing.assert_allclose, out, expected_out)

  def test_graphnetwork_attention_error(self):
    with self.assertRaisesRegex(
        ValueError, ('attention_logit_fn and attention_reduce_fn '
                     'must both be supplied.')):
      models.GraphNetwork(update_edge_fn=None, update_node_fn=None,
                          attention_logit_fn=lambda x: x,
                          attention_reduce_fn=None)
    with self.assertRaisesRegex(
        ValueError, ('attention_logit_fn and attention_reduce_fn '
                     'must both be supplied.')):
      models.GraphNetwork(update_edge_fn=None, update_node_fn=None,
                          attention_logit_fn=None,
                          attention_reduce_fn=lambda x: x)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 DeepMind Technologies Limited.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for sharded graphnet."""

import functools
import os

from absl.testing import absltest
from absl.testing import parameterized
import jax
from jax.lib import xla_bridge
import jax.tree_util as tree
import jraph
from jraph._src import utils
from jraph.experimental import sharded_graphnet
import numpy as np


def _get_graphs_from_n_edge(n_edge):
  """Get a graphs tuple from n_edge."""
  graphs = []
  for el in n_edge:
    graphs.append(
        jraph.GraphsTuple(
            nodes=np.random.uniform(size=(128, 2)),
            edges=np.random.uniform(size=(el, 2)),
            senders=np.random.choice(128, el),
            receivers=np.random.choice(128, el),
            n_edge=np.array([el]),
            n_node=np.array([128]),
            globals=np.array([[el]]),
        ))
  graphs = utils.batch_np(graphs)
  return graphs


def get_graphs_tuples(n_edge, sharded_n_edge, device_graph_idx):
  sharded_n_edge = np.array(sharded_n_edge)
  device_graph_idx = np.array(device_graph_idx)
  devices = len(sharded_n_edge)
  graphs = _get_graphs_from_n_edge(n_edge)
  sharded_senders = np.reshape(graphs.senders, [devices, -1])
  sharded_receivers = np.reshape(graphs.receivers, [devices, -1])
  sharded_edges = np.reshape(graphs.edges,
                             [devices, -1, graphs.edges.shape[-1]])
  # Broadcast replicated features to have a devices leading axis.
  broadcast = lambda x: np.broadcast_to(x[None, :], [devices] + list(x.shape))

  sharded_graphs = sharded_graphnet.ShardedEdgesGraphsTuple(
      device_senders=sharded_senders,
      device_receivers=sharded_receivers,
      device_edges=sharded_edges,
      device_n_edge=sharded_n_edge,
      nodes=broadcast(graphs.nodes),
      senders=broadcast(graphs.senders),
      receivers=broadcast(graphs.receivers),
      device_graph_idx=device_graph_idx,
      globals=broadcast(graphs.globals),
      n_node=broadcast(graphs.n_node),
      n_edge=broadcast(graphs.n_edge))
  return graphs, sharded_graphs


class ShardedGraphnetTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    os.environ[
        'XLA_FLAGS'] = '--xla_force_host_platform_device_count=3'
    xla_bridge.get_backend.cache_clear()

  @parameterized.named_parameters(
      ('split_3_to_4', [3, 5, 4], [[3, 3], [2, 4]], [[0, 1], [1, 2]]),
      ('split_zero_last_edge', [1, 2, 5, 4], [[1, 2, 3], [2, 4, 0]
                                             ], [[0, 1, 2], [2, 3, 0]]),
      ('split_one_over_multiple', [1, 11], [[1, 3], [4, 0], [4, 0]
                                           ], [[0, 1], [1, 0], [1, 0]]))
  def test_get_sharded_graphs_tuple(self, n_edge, sharded_n_edge,
                                    device_graph_idx):
    in_tuple, expect_tuple = get_graphs_tuples(n_edge, sharded_n_edge,
                                               device_graph_idx)
    out_tuple = sharded_graphnet.graphs_tuple_to_broadcasted_sharded_graphs_tuple(
        in_tuple, num_shards=len(expect_tuple.nodes))
    tree.tree_map(np.testing.assert_almost_equal, out_tuple, expect_tuple)

  @parameterized.named_parameters(
      ('split_intermediate', [3, 5, 4, 3, 3]),
      ('split_zero_last_edge', [1, 2, 5, 4, 6]),
      ('split_one_over_multiple', [1, 11]))
  def test_sharded_same_as_non_sharded(self, n_edge):
    in_tuple = _get_graphs_from_n_edge(n_edge)
    devices = 3
    sharded_tuple = sharded_graphnet.graphs_tuple_to_broadcasted_sharded_graphs_tuple(
        in_tuple, devices)
    update_fn = jraph.concatenated_args(lambda x: x)
    sharded_gn = sharded_graphnet.ShardedEdgesGraphNetwork(
        update_fn, update_fn, update_fn, num_shards=devices)
    gn = jraph.GraphNetwork(update_fn, update_fn, update_fn)
    sharded_out = jax.pmap(sharded_gn, axis_name='i')(sharded_tuple)
    expected_out = gn(in_tuple)
    reduced_out = sharded_graphnet.broadcasted_sharded_graphs_tuple_to_graphs_tuple(
        sharded_out)
    jax.tree_util.tree_map(
        functools.partial(np.testing.assert_allclose, atol=1E-5, rtol=1E-5),
        expected_out, reduced_out)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Install script for setuptools."""

from setuptools import find_namespace_packages
from setuptools import setup


def _get_version():
  with open('jraph/__init__.py') as fp:
    for line in fp:
      if line.startswith('__version__') and '=' in line:
        version = line[line.find('=')+1:].strip(' \'"\n')
        if version:
          return version
    raise ValueError('`__version__` not defined in `jraph/__init__.py`')


setup(
    name='jraph',
    version=_get_version(),
    url='https://github.com/deepmind/jraph',
    license='Apache 2.0',
    author='DeepMind',
    description=('Jraph: A library for Graph Neural Networks in Jax'),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author_email='jax_graph_nets@google.com',
    keywords='jax graph neural networks python machine learning',
    packages=find_namespace_packages(exclude=['*_test.py']),
    package_data={'jraph': ['ogb_examples/test_data/*']},
    python_requires='>=3.6',
    install_requires=[
        'jax>=0.1.55',
        'jaxlib>=0.1.37',
        'numpy>=1.18.0',
    ],
    extras_require={'examples': ['dm-haiku>=0.0.2', 'absl-py>=0.9',
                                 'frozendict>=2.0.2', 'optax>=0.0.1',
                                 'scipy>=1.2.1'],
                    'ogb_examples': ['dm-haiku>=0.0.2', 'absl-py>=0.9',
                                     'optax>=0.0.1', 'pandas>=1.0.5',
                                     'dm-tree>=0.1.5']},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# Copyright 2020 DeepMind Technologies Limited.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for working with `GraphsTuple`s."""

import functools
from typing import Any, Callable, Generator, Iterable, Iterator, List, Mapping, Optional, Sequence, Union

import jax
from jax import lax
import jax.numpy as jnp
import jax.tree_util as tree
from jraph._src import graph as gn_graph
import numpy as np

# As of 04/2020 pytype doesn't support recursive types.
# pytype: disable=not-supported-yet
ArrayTree = Union[jnp.ndarray, Iterable['ArrayTree'], Mapping[Any, 'ArrayTree']]


def segment_sum(data: jnp.ndarray,
                segment_ids: jnp.ndarray,
                num_segments: Optional[int] = None,
                indices_are_sorted: bool = False,
                unique_indices: bool = False):
  """Computes the sum within segments of an array.

  Jraph alias to `jax.ops.segment_sum
  <https://jax.readthedocs.io/en/latest/_autosummary/jax.ops.segment_sum.html>`_.
  Note that other segment operations in jraph are not aliases, but are rather
  defined inside the package. Similar to TensorFlow's `segment_sum
  <https://www.tensorflow.org/api_docs/python/tf/math/segment_sum>`_.

  Args:
    data: an array with the values to be summed.
    segment_ids: an array with integer dtype that indicates the segments of
      `data` (along its leading axis) to be summed. Values can be repeated and
      need not be sorted. Values outside of the range [0, num_segments) are
      dropped and do not contribute to the sum.
    num_segments: optional, an int with nonnegative value indicating the number
      of segments. The default is set to be the minimum number of segments that
      would support all indices in ``segment_ids``, calculated as
      ``max(segment_ids) + 1``. Since `num_segments` determines the size of the
      output, a static value must be provided to use ``segment_sum`` in a
      ``jit``-compiled function.
    indices_are_sorted: whether ``segment_ids`` is known to be sorted.
    unique_indices: whether ``segment_ids`` is known to be free of duplicates.

  Returns:
    An array with shape :code:`(num_segments,) + data.shape[1:]` representing
    the segment sums.

  Examples:
    Simple 1D segment sum:

    >>> data = jnp.arange(5)
    >>> segment_ids = jnp.array([0, 0, 1, 1, 2])
    >>> segment_sum(data, segment_ids)
    DeviceArray([1, 5, 4], dtype=int32)

    Using JIT requires static `num_segments`:

    >>> from jax import jit
    >>> jit(segment_sum, static_argnums=2)(data, segment_ids, 3)
    DeviceArray([1, 5, 4], dtype=int32)
  """
  return jax.ops.segment_sum(
      data=data,
      segment_ids=segment_ids,
      num_segments=num_segments,
      indices_are_sorted=indices_are_sorted,
      unique_indices=unique_indices)


def segment_mean(data: jnp.ndarray,
                 segment_ids: jnp.ndarray,
                 num_segments: Optional[int] = None,
                 indices_are_sorted: bool = False,
                 unique_indices: bool = False):
  """Returns mean for each segment.

  Args:
    data: the values which are averaged segment-wise.
    segment_ids: indices for the segments.
    num_segments: total number of segments.
    indices_are_sorted: whether ``segment_ids`` is known to be sorted.
    unique_indices: whether ``segment_ids`` is known to be free of duplicates.
  """
  nominator = segment_sum(
      data,
      segment_ids,
      num_segments,
      indices_are_sorted=indices_are_sorted,
      unique_indices=unique_indices)
  denominator = segment_sum(
      jnp.ones_like(data),
      segment_ids,
      num_segments,
      indices_are_sorted=indices_are_sorted,
      unique_indices=unique_indices)
  return nominator / jnp.maximum(denominator,
                                 jnp.ones(shape=[], dtype=denominator.dtype))


def segment_variance(data: jnp.ndarray,
                     segment_ids: jnp.ndarray,
                     num_segments: Optional[int] = None,
                     indices_are_sorted: bool = False,
                     unique_indices: bool = False):
  """Returns the variance for each segment.

  Args:
    data: values whose variance will be calculated segment-wise.
    segment_ids: indices for segments
    num_segments: total number of segments.
    indices_are_sorted: whether ``segment_ids`` is known to be sorted.
    unique_indices: whether ``segment_ids`` is known to be free of duplicates.

  Returns:
    num_segments size array containing the variance of each segment.
  """
  means = segment_mean(
      data,
      segment_ids,
      num_segments,
      indices_are_sorted=indices_are_sorted,
      unique_indices=unique_indices)[segment_ids]
  counts = segment_sum(
      jnp.ones_like(data),
      segment_ids,
      num_segments,
      indices_are_sorted=indices_are_sorted,
      unique_indices=unique_indices)
  counts = jnp.maximum(counts, jnp.ones_like(counts))
  variances = segment_sum(
      jnp.power(data - means, 2),
      segment_ids,
      num_segments,
      indices_are_sorted=indices_are_sorted,
      unique_indices=unique_indices) / counts
  return variances


def segment_normalize(data: jnp.ndarray,
                      segment_ids: jnp.ndarray,
                      num_segments: Optional[int] = None,
                      indices_are_sorted: bool = False,
                      unique_indices: bool = False,
                      eps=1e-8):
  """Normalizes data within each segment.

  Args:
    data: values whose z-score normalized values will be calculated.
      segment-wise.
    segment_ids: indices for segments.
    num_segments: total number of segments.
    indices_are_sorted: whether ``segment_ids`` is known to be sorted.
    unique_indices: whether ``segment_ids`` is known to be free of duplicates.
    eps: epsilon for numerical stability.

  Returns:
    array containing data normalized segment-wise.
  """

  means = segment_mean(
      data,
      segment_ids,
      num_segments,
      indices_are_sorted=indices_are_sorted,
      unique_indices=unique_indices)[segment_ids]
  variances = segment_variance(
      data,
      segment_ids,
      num_segments,
      indices_are_sorted=indices_are_sorted,
      unique_indices=unique_indices)[segment_ids]
  normalized = (data - means) * lax.rsqrt(jnp.maximum(
      variances, jnp.array(eps, dtype=variances.dtype)))
  return normalized


def segment_max(data: jnp.ndarray,
                segment_ids: jnp.ndarray,
                num_segments: Optional[int] = None,
                indices_are_sorted: bool = False,
                unique_indices: bool = False):
  """Alias for jax.ops.segment_max.

  Args:
    data: an array with the values to be maxed over.
    segment_ids: an array with integer dtype that indicates the segments of
      `data` (along its leading axis) to be maxed over. Values can be repeated
      and need not be sorted. Values outside of the range [0, num_segments) are
      dropped and do not contribute to the result.
    num_segments: optional, an int with positive value indicating the number of
      segments. The default is ``jnp.maximum(jnp.max(segment_ids) + 1,
      jnp.max(-segment_ids))`` but since `num_segments` determines the size of
      the output, a static value must be provided to use ``segment_max`` in a
      ``jit``-compiled function.
    indices_are_sorted: whether ``segment_ids`` is known to be sorted
    unique_indices: whether ``segment_ids`` is known to be free of duplicates

  Returns:
    An array with shape ``(num_segments,) + data.shape[1:]`` representing
    the segment maxs.
  """
  return jax.ops.segment_max(data, segment_ids, num_segments,
                             indices_are_sorted, unique_indices)


def _replace_empty_segments_with_constant(aggregated_segments: jnp.ndarray,
                                          segment_ids: jnp.ndarray,
                                          num_segments: Optional[int] = None,
                                          constant: float = 0.):
  """Replaces the values of empty segments with constants."""
  result_shape = (len(segment_ids),) + aggregated_segments.shape[1:]
  num_elements_in_segment = segment_sum(
      jnp.ones(result_shape, dtype=jnp.int32),
      segment_ids,
      num_segments=num_segments)
  return jnp.where(num_elements_in_segment > 0, aggregated_segments,
                   jnp.array(constant, dtype=aggregated_segments.dtype))


def segment_min_or_constant(data: jnp.ndarray,
                            segment_ids: jnp.ndarray,
                            num_segments: Optional[int] = None,
                            indices_are_sorted: bool = False,
                            unique_indices: bool = False,
                            constant: float = 0.):
  """As segment_min, but returns a constant for empty segments.

  `segment_min` returns `-inf` for empty segments, which can cause `nan`s in the
  backwards pass of a neural network, even with masking. This method overrides
  the default behaviour of `segment_min` and returns a constant for empty
  segments.

  Args:
    data: an array with the values to be maxed over.
    segment_ids: an array with integer dtype that indicates the segments of
      `data` (along its leading axis) to be maxed over. Values can be repeated
      and need not be sorted. Values outside of the range [0, num_segments) are
      dropped and do not contribute to the result.
    num_segments: optional, an int with positive value indicating the number of
      segments. The default is ``jnp.maximum(jnp.max(segment_ids) + 1,
      jnp.max(-segment_ids))`` but since `num_segments` determines the size of
      the output, a static value must be provided to use ``segment_min`` in a
      ``jit``-compiled function.
    indices_are_sorted: whether ``segment_ids`` is known to be sorted
    unique_indices: whether ``segment_ids`` is known to be free of duplicates
    constant: The constant to replace empty segments with, defaults to zero.

  Returns:
    An array with shape ``(num_segments,) + data.shape[1:]`` representing
    the segment maxs.
  """
  mins_ = segment_min(data, segment_ids, num_segments, indices_are_sorted,
                      unique_indices)
  return _replace_empty_segments_with_constant(mins_, segment_ids, num_segments,
                                               constant)


def segment_max_or_constant(data: jnp.ndarray,
                            segment_ids: jnp.ndarray,
                            num_segments: Optional[int] = None,
                            indices_are_sorted: bool = False,
                            unique_indices: bool = False,
                            constant: float = 0.):
  """As segment_max, but returns a constant for empty segments.

  `segment_max` returns `-inf` for empty segments, which can cause `nan`s in the
  backwards pass of a neural network, even with masking. This method overrides
  the default behaviour of `segment_max` and returns a constant for empty
  segments.

  Args:
    data: an array with the values to be maxed over.
    segment_ids: an array with integer dtype that indicates the segments of
      `data` (along its leading axis) to be maxed over. Values can be repeated
      and need not be sorted. Values outside of the range [0, num_segments) are
      dropped and do not contribute to the result.
    num_segments: optional, an int with positive value indicating the number of
      segments. The default is ``jnp.maximum(jnp.max(segment_ids) + 1,
      jnp.max(-segment_ids))`` but since `num_segments` determines the size of
      the output, a static value must be provided to use ``segment_max`` in a
      ``jit``-compiled function.
    indices_are_sorted: whether ``segment_ids`` is known to be sorted
    unique_indices: whether ``segment_ids`` is known to be free of duplicates
    constant: The constant to replace empty segments with, defaults to zero.

  Returns:
    An array with shape ``(num_segments,) + data.shape[1:]`` representing
    the segment maxs.
  """
  maxs_ = segment_max(data, segment_ids, num_segments, indices_are_sorted,
                      unique_indices)
  return _replace_empty_segments_with_constant(maxs_, segment_ids, num_segments,
                                               constant)


def segment_min(data: jnp.ndarray,
                segment_ids: jnp.ndarray,
                num_segments: Optional[int] = None,
                indices_are_sorted: bool = False,
                unique_indices: bool = False):
  """Computes the min within segments of an array.

  Similar to TensorFlow's segment_min:
  https://www.tensorflow.org/api_docs/python/tf/math/segment_min

  Args:
    data: an array with the values to be maxed over.
    segment_ids: an array with integer dtype that indicates the segments of
      `data` (along its leading axis) to be min'd over. Values can be repeated
      and need not be sorted. Values outside of the range [0, num_segments) are
      dropped and do not contribute to the result.
    num_segments: optional, an int with positive value indicating the number of
      segments. The default is ``jnp.maximum(jnp.max(segment_ids) + 1,
      jnp.max(-segment_ids))`` but since `num_segments` determines the size of
      the output, a static value must be provided to use ``segment_max`` in a
      ``jit``-compiled function.
    indices_are_sorted: whether ``segment_ids`` is known to be sorted
    unique_indices: whether ``segment_ids`` is known to be free of duplicates

  Returns:
    An array with shape ``(num_segments,) + data.shape[1:]`` representing
    the segment mins.
  """
  return jax.ops.segment_min(data, segment_ids, num_segments,
                             indices_are_sorted, unique_indices)


def segment_softmax(logits: jnp.ndarray,
                    segment_ids: jnp.ndarray,
                    num_segments: Optional[int] = None,
                    indices_are_sorted: bool = False,
                    unique_indices: bool = False) -> ArrayTree:
  """Computes a segment-wise softmax.

  For a given tree of logits that can be divded into segments, computes a
  softmax over the segments.

    logits = jnp.ndarray([1.0, 2.0, 3.0, 1.0, 2.0])
    segment_ids = jnp.ndarray([0, 0, 0, 1, 1])
    segment_softmax(logits, segments)
    >> DeviceArray([0.09003057, 0.24472848, 0.66524094, 0.26894142, 0.7310586],
    >> dtype=float32)

  Args:
    logits: an array of logits to be segment softmaxed.
    segment_ids: an array with integer dtype that indicates the segments of
      `data` (along its leading axis) to be maxed over. Values can be repeated
      and need not be sorted. Values outside of the range [0, num_segments) are
      dropped and do not contribute to the result.
    num_segments: optional, an int with positive value indicating the number of
      segments. The default is ``jnp.maximum(jnp.max(segment_ids) + 1,
      jnp.max(-segment_ids))`` but since ``num_segments`` determines the size of
      the output, a static value must be provided to use ``segment_sum`` in a
      ``jit``-compiled function.
    indices_are_sorted: whether ``segment_ids`` is known to be sorted
    unique_indices: whether ``segment_ids`` is known to be free of duplicates

  Returns:
    The segment softmax-ed ``logits``.
  """
  # First, subtract the segment max for numerical stability
  maxs = segment_max(logits, segment_ids, num_segments, indices_are_sorted,
                     unique_indices)
  logits = logits - maxs[segment_ids]
  # Then take the exp
  logits = jnp.exp(logits)
  # Then calculate the normalizers
  normalizers = segment_sum(logits, segment_ids, num_segments,
                            indices_are_sorted, unique_indices)
  normalizers = normalizers[segment_ids]
  softmax = logits / normalizers
  return softmax


def partition_softmax(logits: ArrayTree,
                      partitions: jnp.ndarray,
                      sum_partitions: Optional[int] = None):
  """Compute a softmax within partitions of an array.

    For example::
      logits = jnp.ndarray([1.0, 2.0, 3.0, 1.0, 2.0])
      partitions = jnp.ndarray([3, 2])
      partition_softmax(node_logits, n_node)
      >> DeviceArray(
      >> [0.09003057, 0.24472848, 0.66524094, 0.26894142, 0.7310586],
      >> dtype=float32)

  Args:
    logits: the logits for the softmax.
    partitions: the number of nodes per graph. It is a vector of integers with
      shape ``[n_graphs]``, such that ``graph.n_node[i]`` is the number of nodes
      in the i-th graph.
    sum_partitions: the sum of n_node. If not passed, the result of this method
      is data dependent and so not ``jit``-able.

  Returns:
    The softmax over partitions.
  """
  n_partitions = len(partitions)
  segment_ids = jnp.repeat(
      jnp.arange(n_partitions),
      partitions,
      axis=0,
      total_repeat_length=sum_partitions)
  return segment_softmax(
      logits, segment_ids, n_partitions, indices_are_sorted=True)


def batch(graphs: Sequence[gn_graph.GraphsTuple]) -> gn_graph.GraphsTuple:
  """Returns a batched graph given a list of graphs.

  This method will concatenate the ``nodes``, ``edges`` and ``globals``,
  ``n_node`` and ``n_edge`` of a sequence of ``GraphsTuple`` along axis 0. For
  ``senders`` and ``receivers``, offsets are computed so that connectivity
  remains valid for the new node indices.

  For example::

    key = jax.random.PRNGKey(0)
    graph_1 = GraphsTuple(nodes=jax.random.normal(key, (3, 64)),
                      edges=jax.random.normal(key, (5, 64)),
                      senders=jnp.array([0,0,1,1,2]),
                      receivers=[1,2,0,2,1],
                      n_node=jnp.array([3]),
                      n_edge=jnp.array([5]),
                      globals=jax.random.normal(key, (1, 64)))
    graph_2 = GraphsTuple(nodes=jax.random.normal(key, (5, 64)),
                      edges=jax.random.normal(key, (10, 64)),
                      senders=jnp.array([0,0,1,1,2,2,3,3,4,4]),
                      receivers=jnp.array([1,2,0,2,1,0,2,1,3,2]),
                      n_node=jnp.array([5]),
                      n_edge=jnp.array([10]),
                      globals=jax.random.normal(key, (1, 64)))
    batch = graph.batch([graph_1, graph_2])

    batch.nodes.shape
    >> (8, 64)
    batch.edges.shape
    >> (15, 64)
    # Offsets computed on senders and receivers
    batch.senders
    >> DeviceArray([0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7], dtype=int32)
    batch.receivers
    >> DeviceArray([1, 2, 0, 2, 1, 4, 5, 3, 5, 4, 3, 5, 4, 6, 5], dtype=int32)
    batch.n_node
    >> DeviceArray([3, 5], dtype=int32)
    batch.n_edge
    >> DeviceArray([5, 10], dtype=int32)

  If a ``GraphsTuple`` does not contain any graphs, it will be dropped from the
  batch.

  This method is not compilable because it is data dependent.

  This functionality was implementation as  ``utils_tf.concat`` in the
  Tensorflow version of graph_nets.

  Args:
    graphs: sequence of ``GraphsTuple``s which will be batched into a single
      graph.
  """
  return _batch(graphs, np_=jnp)


def batch_np(graphs: Sequence[gn_graph.GraphsTuple]) -> gn_graph.GraphsTuple:
  """Numpy implementation of `batch`. See `batch` for more details."""
  return _batch(graphs, np_=np)


def _batch(graphs, np_):
  """Returns batched graph given a list of graphs and a numpy-like module."""
  # Calculates offsets for sender and receiver arrays, caused by concatenating
  # the nodes arrays.
  offsets = np_.cumsum(
      np_.array([0] + [np_.sum(g.n_node) for g in graphs[:-1]]))

  def _map_concat(nests):
    concat = lambda *args: np_.concatenate(args)
    return tree.tree_map(concat, *nests)

  return gn_graph.GraphsTuple(
      n_node=np_.concatenate([g.n_node for g in graphs]),
      n_edge=np_.concatenate([g.n_edge for g in graphs]),
      nodes=_map_concat([g.nodes for g in graphs]),
      edges=_map_concat([g.edges for g in graphs]),
      globals=_map_concat([g.globals for g in graphs]),
      senders=np_.concatenate([g.senders + o for g, o in zip(graphs, offsets)]),
      receivers=np_.concatenate(
          [g.receivers + o for g, o in zip(graphs, offsets)]))


def unbatch(graph: gn_graph.GraphsTuple) -> List[gn_graph.GraphsTuple]:
  """Returns a list of graphs given a batched graph.

  This function does not support jax.jit, because the shape of the output
  is data-dependent!

  Args:
    graph: the batched graph, which will be unbatched into a list of graphs.
  """
  return _unbatch(graph, np_=jnp)


def unbatch_np(graph: gn_graph.GraphsTuple) -> List[gn_graph.GraphsTuple]:
  """Numpy implementation of `unbatch`. See `unbatch` for more details."""
  return _unbatch(graph, np_=np)


def _unbatch(graph: gn_graph.GraphsTuple, np_) -> List[gn_graph.GraphsTuple]:
  """Returns a list of graphs given a batched graph."""

  def _map_split(nest, indices_or_sections):
    """Splits leaf nodes of nests and returns a list of nests."""
    if isinstance(indices_or_sections, int):
      n_lists = indices_or_sections
    else:
      n_lists = len(indices_or_sections) + 1
    concat = lambda field: np_.split(field, indices_or_sections)
    nest_of_lists = tree.tree_map(concat, nest)
    # pylint: disable=cell-var-from-loop
    list_of_nests = [
        tree.tree_map(lambda _, x: x[i], nest, nest_of_lists)
        for i in range(n_lists)
    ]
    return list_of_nests

  all_n_node = graph.n_node[:, None]
  all_n_edge = graph.n_edge[:, None]
  node_offsets = np_.cumsum(graph.n_node[:-1])
  all_nodes = _map_split(graph.nodes, node_offsets)
  edge_offsets = np_.cumsum(graph.n_edge[:-1])
  all_edges = _map_split(graph.edges, edge_offsets)
  all_globals = _map_split(graph.globals, len(graph.n_node))
  all_senders = np_.split(graph.senders, edge_offsets)
  all_receivers = np_.split(graph.receivers, edge_offsets)

  # Corrects offset in the sender and receiver arrays, caused by splitting the
  # nodes array.
  n_graphs = graph.n_node.shape[0]
  for graph_index in np_.arange(n_graphs)[1:]:
    all_senders[graph_index] -= node_offsets[graph_index - 1]
    all_receivers[graph_index] -= node_offsets[graph_index - 1]

  return [
      gn_graph.GraphsTuple._make(elements)
      for elements in zip(all_nodes, all_edges, all_receivers, all_senders,
                          all_globals, all_n_node, all_n_edge)
  ]


def pad_with_graphs(graph: gn_graph.GraphsTuple,
                    n_node: int,
                    n_edge: int,
                    n_graph: int = 2) -> gn_graph.GraphsTuple:
  """Pads a ``GraphsTuple`` to size by adding computation preserving graphs.

  The ``GraphsTuple`` is padded by first adding a dummy graph which contains the
  padding nodes and edges, and then empty graphs without nodes or edges.

  The empty graphs and the dummy graph do not interfer with the graphnet
  calculations on the original graph, and so are computation preserving.

  The padding graph requires at least one node and one graph.

  This function does not support jax.jit, because the shape of the output
  is data-dependent.

  Args:
    graph: ``GraphsTuple`` padded with dummy graph and empty graphs.
    n_node: the number of nodes in the padded ``GraphsTuple``.
    n_edge: the number of edges in the padded ``GraphsTuple``.
    n_graph: the number of graphs in the padded ``GraphsTuple``. Default is 2,
      which is the lowest possible value, because we always have at least one
      graph in the original ``GraphsTuple`` and we need one dummy graph for the
      padding.

  Raises:
    ValueError: if the passed ``n_graph`` is smaller than 2.
    RuntimeError: if the given ``GraphsTuple`` is too large for the given
      padding.

  Returns:
    A padded ``GraphsTuple``.
  """
  if n_graph < 2:
    raise ValueError(
        f'n_graph is {n_graph}, which is smaller than minimum value of 2.')
  graph = jax.device_get(graph)
  pad_n_node = int(n_node - np.sum(graph.n_node))
  pad_n_edge = int(n_edge - np.sum(graph.n_edge))
  pad_n_graph = int(n_graph - graph.n_node.shape[0])
  if pad_n_node <= 0 or pad_n_edge < 0 or pad_n_graph <= 0:
    raise RuntimeError(
        'Given graph is too large for the given padding. difference: '
        f'n_node {pad_n_node}, n_edge {pad_n_edge}, n_graph {pad_n_graph}')

  pad_n_empty_graph = pad_n_graph - 1

  tree_nodes_pad = (
      lambda leaf: np.zeros((pad_n_node,) + leaf.shape[1:], dtype=leaf.dtype))
  tree_edges_pad = (
      lambda leaf: np.zeros((pad_n_edge,) + leaf.shape[1:], dtype=leaf.dtype))
  tree_globs_pad = (
      lambda leaf: np.zeros((pad_n_graph,) + leaf.shape[1:], dtype=leaf.dtype))

  padding_graph = gn_graph.GraphsTuple(
      n_node=np.concatenate(
          [np.array([pad_n_node], dtype=np.int32),
           np.zeros(pad_n_empty_graph, dtype=np.int32)]),
      n_edge=np.concatenate(
          [np.array([pad_n_edge], dtype=np.int32),
           np.zeros(pad_n_empty_graph, dtype=np.int32)]),
      nodes=tree.tree_map(tree_nodes_pad, graph.nodes),
      edges=tree.tree_map(tree_edges_pad, graph.edges),
      globals=tree.tree_map(tree_globs_pad, graph.globals),
      senders=np.zeros(pad_n_edge, dtype=np.int32),
      receivers=np.zeros(pad_n_edge, dtype=np.int32),
  )
  return _batch([graph, padding_graph], np_=np)


def get_number_of_padding_with_graphs_graphs(
    padded_graph: gn_graph.GraphsTuple) -> int:
  """Returns number of padding graphs in padded_graph.

  Warning: This method only gives results for graphs that have been padded with
  ``pad_with_graphs``.

  Args:
    padded_graph: a ``GraphsTuple`` that has been padded with
      ``pad_with_graphs``.

  Returns:
    The number of padding graphs.
  """
  # The first_padding graph always has at least one padding node, and
  # all padding graphs that follow have 0 nodes. We can count how many
  # trailing graphs have 0 nodes, and add one.
  n_trailing_empty_padding_graphs = jnp.argmin(padded_graph.n_node[::-1] == 0)
  return n_trailing_empty_padding_graphs + 1


def get_number_of_padding_with_graphs_nodes(
    padded_graph: gn_graph.GraphsTuple) -> int:
  """Returns number of padding nodes in given padded_graph.

  Warning: This method only gives results for graphs that have been padded with
  ``pad_with_graphs``.

  Args:
    padded_graph: a ``GraphsTuple`` that has been padded with
      ``pad_with_graphs``.

  Returns:
    The number of padding nodes.
  """
  return padded_graph.n_node[
      -get_number_of_padding_with_graphs_graphs(padded_graph)]


def get_number_of_padding_with_graphs_edges(
    padded_graph: gn_graph.GraphsTuple) -> int:
  """Returns number of padding edges in given padded_graph.

  Warning: This method only gives results for graphs that have been padded with
  ``pad_with_graphs``.

  Args:
    padded_graph: a ``GraphsTuple`` that has been padded with
      ``pad_with_graphs``.

  Returns:
    The number of padding edges.
  """
  return padded_graph.n_edge[
      -get_number_of_padding_with_graphs_graphs(padded_graph)]


def unpad_with_graphs(
    padded_graph: gn_graph.GraphsTuple) -> gn_graph.GraphsTuple:
  """Unpads the given graph by removing the dummy graph and empty graphs.

  This function assumes that the given graph was padded with the
  ``pad_with_graphs`` function.

  This function does not support jax.jit, because the shape of the output
  is data-dependent!

  Args:
    padded_graph: ``GraphsTuple`` padded with a dummy graph and empty graphs.

  Returns:
    The unpadded graph.
  """
  n_padding_graph = get_number_of_padding_with_graphs_graphs(padded_graph)
  n_padding_node = get_number_of_padding_with_graphs_nodes(padded_graph)
  n_padding_edge = get_number_of_padding_with_graphs_edges(padded_graph)

  def remove_edge_padding(edge_array):
    if n_padding_edge == 0:
      return edge_array
    return edge_array[:-n_padding_edge]

  unpadded_graph = gn_graph.GraphsTuple(
      n_node=padded_graph.n_node[:-n_padding_graph],
      n_edge=padded_graph.n_edge[:-n_padding_graph],
      nodes=tree.tree_map(lambda x: x[:-n_padding_node], padded_graph.nodes),
      edges=tree.tree_map(remove_edge_padding, padded_graph.edges),
      globals=tree.tree_map(lambda x: x[:-n_padding_graph],
                            padded_graph.globals),
      senders=remove_edge_padding(padded_graph.senders),
      receivers=remove_edge_padding(padded_graph.receivers),
  )
  return unpadded_graph


def get_node_padding_mask(padded_graph: gn_graph.GraphsTuple) -> ArrayTree:
  """Returns a mask for the nodes of a padded graph.

  Args:
    padded_graph: ``GraphsTuple`` padded using ``pad_with_graphs``. This graph
      must contain at least one array of node features so the total static
      number of nodes can be inferred statically from the shape, and the method
      can be jitted.

  Returns:
    Boolean array of shape [total_num_nodes] containing True for real nodes,
    and False for padding nodes.
  """
  n_padding_node = get_number_of_padding_with_graphs_nodes(padded_graph)
  flat_node_features = tree.tree_leaves(padded_graph.nodes)

  if not flat_node_features:
    raise ValueError(
        '`padded_graph` must have at least one array of node features')
  total_num_nodes = flat_node_features[0].shape[0]
  return _get_mask(padding_length=n_padding_node, full_length=total_num_nodes)


def get_edge_padding_mask(padded_graph: gn_graph.GraphsTuple) -> ArrayTree:
  """Returns a mask for the edges of a padded graph.

  Args:
    padded_graph: ``GraphsTuple`` padded using ``pad_with_graphs``.

  Returns:
    Boolean array of shape [total_num_edges] containing True for real edges,
    and False for padding edges.
  """
  n_padding_edge = get_number_of_padding_with_graphs_edges(padded_graph)
  total_num_edges = padded_graph.senders.shape[0]
  return _get_mask(padding_length=n_padding_edge, full_length=total_num_edges)


def get_graph_padding_mask(padded_graph: gn_graph.GraphsTuple) -> ArrayTree:
  """Returns a mask for the graphs of a padded graph.

  Args:
    padded_graph: ``GraphsTuple`` padded using ``pad_with_graphs``.

  Returns:
    Boolean array of shape [total_num_graphs] containing True for real graphs,
    and False for padding graphs.
  """
  n_padding_graph = get_number_of_padding_with_graphs_graphs(padded_graph)
  total_num_graphs = padded_graph.n_node.shape[0]
  return _get_mask(padding_length=n_padding_graph, full_length=total_num_graphs)


def _get_mask(padding_length, full_length):
  valid_length = full_length - padding_length
  return jnp.arange(full_length, dtype=jnp.int32) < valid_length


def concatenated_args(
    update: Optional[Callable[..., ArrayTree]] = None,
    *,
    axis: int = -1
) -> Union[Callable[..., ArrayTree], Callable[[Callable[..., ArrayTree]],
                                              ArrayTree]]:
  """Decorator that concatenates arguments before being passed to an update_fn.

  By default node, edge and global features are passed separately to update
  functions. However, it is common practice to concatenate these features before
  passing them to a neural network. This wrapper concatenates the arguments
  for you.

  For example::

    # Without the wrapper
    def update_node_fn(nodes, receivers, globals):
      return net(jnp.concatenate([nodes, receivers, globals], axis=1))

    # With the wrapper
    @concatenated_args
    def update_node_fn(features):
      return net(features)

  Args:
    update: an update function that takes ``jnp.ndarray``.
    axis: the axis upon which to concatenate.

  Returns:
    A wrapped function with the arguments concatenated.
  """

  def _decorate(f):

    @functools.wraps(update)
    def wrapper(*args, **kwargs):
      combined_args = tree.tree_flatten(args)[0] + tree.tree_flatten(kwargs)[0]
      concat_args = jnp.concatenate(combined_args, axis=axis)
      return f(concat_args)

    return wrapper

  # If the update function is passed, then decorate the update function.
  if update:
    return _decorate(update)

  # Otherwise, return the decorator.
  return _decorate


def dtype_max_value(dtype):
  if dtype.kind == 'f':
    return jnp.inf
  elif dtype.kind == 'i':
    return jnp.iinfo(dtype).max
  elif dtype.kind == 'b':
    return True
  else:
    raise ValueError(f'Invalid data type {dtype.kind!r}.')


def dtype_min_value(dtype):
  if dtype.kind == 'f':
    return -jnp.inf
  elif dtype.kind == 'i':
    return jnp.iinfo(dtype).min
  elif dtype.kind == 'b':
    return False
  else:
    raise ValueError(f'Invalid data type {dtype.kind!r}.')


def get_fully_connected_graph(n_node_per_graph: int,
                              n_graph: int,
                              node_features: Optional[ArrayTree] = None,
                              global_features: Optional[ArrayTree] = None,
                              add_self_edges: bool = True):
  """Gets a fully connected graph given n_node_per_graph and n_graph.

  This method is jittable.

  Args:
    n_node_per_graph: The number of nodes in each graph.
    n_graph: The number of graphs in the `jraph.GraphsTuple`.
    node_features: Optional node features.
    global_features: Optional global features.
    add_self_edges: Whether to add self edges to the graph.

  Returns:
    `jraph.GraphsTuple`
  """
  if node_features is not None:
    num_node_features = jax.tree_leaves(node_features)[0].shape[0]
    if n_node_per_graph * n_graph != num_node_features:
      raise ValueError(
          'Number of nodes is not equal to num_nodes_per_graph * n_graph.')
  if global_features is not None:
    if n_graph != jax.tree_leaves(global_features)[0].shape[0]:
      raise ValueError('The number of globals is not equal to n_graph.')
  senders = []
  receivers = []
  n_edge = []
  tmp_senders, tmp_receivers = jnp.meshgrid(
      jnp.arange(n_node_per_graph), jnp.arange(n_node_per_graph))
  if not add_self_edges:
    tmp_senders = jax.vmap(jnp.roll)(tmp_senders,
                                     -jnp.arange(len(tmp_senders)))[:, 1:]
    tmp_receivers = tmp_receivers[:, 1:]
  # Flatten the senders and receivers.
  tmp_senders = tmp_senders.flatten()
  tmp_receivers = tmp_receivers.flatten()
  for graph_idx in range(n_graph):
    offset = graph_idx * n_node_per_graph
    senders.append(tmp_senders + offset)
    receivers.append(tmp_receivers + offset)
    n_edge.append(len(tmp_senders))

  def _concat_or_empty_indices(indices_list):
    if indices_list:
      return jnp.concatenate(indices_list, axis=0)
    else:
      return jnp.array([], dtype=tmp_senders.dtype)

  return gn_graph.GraphsTuple(
      nodes=node_features,
      edges=None,
      n_node=jnp.array([n_node_per_graph] * n_graph),
      n_edge=jnp.array(n_edge) if n_edge else jnp.array([0]),
      senders=_concat_or_empty_indices(senders),
      receivers=_concat_or_empty_indices(receivers),
      globals=global_features,
  )


_NUMBER_FIELDS = ('n_node', 'n_edge', 'n_graph')


def _get_graph_size(graphs_tuple):
  n_node = np.sum(graphs_tuple.n_node)
  n_edge = len(graphs_tuple.senders)
  n_graph = len(graphs_tuple.n_node)
  return n_node, n_edge, n_graph


def _is_over_batch_size(graph, graph_batch_size):
  graph_size = _get_graph_size(graph)
  return any([x > y for x, y in zip(graph_size, graph_batch_size)])


def dynamically_batch(
    graphs_tuple_iterator: Iterator[gn_graph.GraphsTuple], n_node: int,
    n_edge: int, n_graph: int) -> Generator[gn_graph.GraphsTuple, None, None]:
  """Dynamically batches trees with `jraph.GraphsTuples` up to specified sizes.


  Elements of the `graphs_tuple_iterator` will be incrementally added to a batch
  until the limits defined by `n_node`, `n_edge` and `n_graph` are reached. This
  means each element yielded by this generator may have a differing number of
  graphs in its batch.

  Args:
    graphs_tuple_iterator: An iterator of `jraph.GraphsTuples`.
    n_node: The maximum number of nodes in a batch, at least the maximum sized
      graph + 1.
    n_edge: The maximum number of edges in a batch, at least the maximum sized
      graph.
    n_graph: The maximum number of graphs in a batch, at least 2.

  Yields:
    A `jraph.GraphsTuple` batch of graphs.

  Raises:
    ValueError: if the number of graphs is < 2.
    RuntimeError: if the `graphs_tuple_iterator` contains elements which are not
      `jraph.GraphsTuple`s.
    RuntimeError: if a graph is found which is larger than the batch size.
  """
  if n_graph < 2:
    raise ValueError('The number of graphs in a batch size must be greater or '
                     f'equal to `2` for padding with graphs, got {n_graph}.')
  valid_batch_size = (n_node - 1, n_edge, n_graph - 1)
  accumulated_graphs = []
  num_accumulated_nodes = 0
  num_accumulated_edges = 0
  num_accumulated_graphs = 0
  for element in graphs_tuple_iterator:
    element_nodes, element_edges, element_graphs = _get_graph_size(element)
    if _is_over_batch_size(element, valid_batch_size):
      # First yield the batched graph so far if exists.
      if accumulated_graphs:
        batched_graph = batch_np(accumulated_graphs)
        yield pad_with_graphs(batched_graph, n_node, n_edge, n_graph)

      # Then report the error.
      graph_size = element_nodes, element_edges, element_graphs
      graph_size = {k: v for k, v in zip(_NUMBER_FIELDS, graph_size)}
      batch_size = {k: v for k, v in zip(_NUMBER_FIELDS, valid_batch_size)}
      raise RuntimeError('Found graph bigger than batch size. Valid Batch '
                         f'Size: {batch_size}, Graph Size: {graph_size}')

    # If this is the first element of the batch, set it and continue.
    # Otherwise check if there is space for the graph in the batch:
    #   if there is, add it to the batch
    #   if there isn't, return the old batch and start a new batch.
    if not accumulated_graphs:
      accumulated_graphs = [element]
      num_accumulated_nodes = element_nodes
      num_accumulated_edges = element_edges
      num_accumulated_graphs = element_graphs
      continue
    else:
      if ((num_accumulated_graphs + element_graphs > n_graph - 1) or
          (num_accumulated_nodes + element_nodes > n_node - 1) or
          (num_accumulated_edges + element_edges > n_edge)):
        batched_graph = batch_np(accumulated_graphs)
        yield pad_with_graphs(batched_graph, n_node, n_edge, n_graph)
        accumulated_graphs = [element]
        num_accumulated_nodes = element_nodes
        num_accumulated_edges = element_edges
        num_accumulated_graphs = element_graphs
      else:
        accumulated_graphs.append(element)
        num_accumulated_nodes += element_nodes
        num_accumulated_edges += element_edges
        num_accumulated_graphs += element_graphs

  # We may still have data in batched graph.
  if accumulated_graphs:
    batched_graph = batch_np(accumulated_graphs)
    yield pad_with_graphs(batched_graph, n_node, n_edge, n_graph)


def _expand_trailing_dimensions(array: jnp.ndarray,
                                template: jnp.ndarray) -> jnp.ndarray:
  missing_dims = len(template.shape) - len(array.shape)
  out = jnp.reshape(array, array.shape + (1,) * missing_dims)
  assert out.dtype == array.dtype
  return out


def _get_zero_fn(mask: jnp.ndarray) -> Callable[[jnp.ndarray], jnp.ndarray]:
  return lambda x: _expand_trailing_dimensions(mask, x) * x


def zero_out_padding(graph: gn_graph.GraphsTuple) -> gn_graph.GraphsTuple:
  """Zeroes out padded graphs values.

  Padded graphs can cause numeric overflow issues when a node has a large number
  connecting edges, and the graph network has many layers. By inserting this
  zeroing function between layers of the graph network these issues may be
  addressed.

  Zeroing is performed by multiplying a boolean mask with the corresponding
  array of the input dtype, assuming correct type casting happens.

  For example::

    # Set up padded jraph.GraphsTuple & simple jraph.GraphNetwork.
    key = jax.random.PRNGKey(0)
    graph = jraph.GraphsTuple(nodes=jax.random.randint(key, (3,), 0, 1,
    dtype=jnp.int32),
                      edges=jax.random.randint(key, (5,), 0, 1,
                      dtype=jnp.int32),
                      senders=jnp.array([0,0,1,1,2]),
                      receivers=[1,2,0,2,1],
                      n_node=jnp.array([3]),
                      n_edge=jnp.array([5]),
                      globals=jax.random.randint(key, (1,), 0, 128,
                      dtype=jnp.int32))
    padded_graph = jraph.pad_with_graphs(graph, n_node=4, n_edge=100, n_graph=2)

    def net_fn(graph: jraph.GraphsTuple) -> jraph.GraphsTuple:
      embedder = jraph.GraphMapFeatures(
          hk.Embed(2, 1), hk.Embed(2, 1), hk.Embed(2, 1))
      update_fn = jraph.concatenated_args(lambda x: jnp.sum(x, -1,
      keepdims=True))
      net = jraph.GraphNetwork(
          update_node_fn=update_fn,
          update_edge_fn=update_fn,
          update_global_fn=update_fn)
      embedded = embedder(graph)
      return net(embedded)

    net = haiku.without_apply_rng(haiku.transform(net_fn))
    params = net.init(key, padded_graph)

    # Without zeroing.
    net.apply(params, padded_graph).nodes
    >> [[ -1.1949954]
        [ -1.4456191]
        [ -1.1949954]
        [-88.505775 ]] # Very large activation in a single node.

    # With zeroing.
    zero_out_padding(net.apply(params, padded_graph)).nodes
    >> [[ -1.1949954]
        [ -1.4456191]
        [ -1.1949954]
        [-0. ]]  # Zeroed out activation.

  Args:
    graph: A padded graph.

  Returns:
    A graph with the same valid values as input, but padded values zeroed out.
  """
  edge_mask = get_edge_padding_mask(graph)
  masked_edges = jax.tree_map(_get_zero_fn(edge_mask), graph.edges)
  node_mask = get_node_padding_mask(graph)
  masked_nodes = jax.tree_map(_get_zero_fn(node_mask), graph.nodes)
  global_mask = get_graph_padding_mask(graph)
  masked_globals = jax.tree_map(_get_zero_fn(global_mask), graph.globals)
  return graph._replace(
      nodes=masked_nodes, edges=masked_edges, globals=masked_globals)


def with_zero_out_padding_outputs(
    graph_net: Callable[[gn_graph.GraphsTuple], gn_graph.GraphsTuple]
) -> Callable[[gn_graph.GraphsTuple], gn_graph.GraphsTuple]:
  """A wrapper for graph to graph functions that zeroes padded d output values.

  See `zero_out_padding` for a full explanation of the method.

  Args:
    graph_net: A Graph Neural Network.

  Returns:
    A Graph Neural Network that will zero out all output padded values.
  """

  @functools.wraps(graph_net)
  def wrapper(graph: gn_graph.GraphsTuple) -> gn_graph.GraphsTuple:
    return zero_out_padding(graph_net(graph))

  return wrapper



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Electronic Voting Example.

In this example we use DeepSets to estimate the winner of an election.
Each vote is represented by a one-hot encoded vector.

It goes without saying, but don't use this in a real election!
Seriously, don't!
"""

import collections
import logging
import random

from absl import app
import haiku as hk
import jax
import jax.numpy as jnp
import jraph
import numpy as np
import optax


Problem = collections.namedtuple("Problem", ("graph", "labels"))


def get_voting_problem(min_n_voters: int, max_n_voters: int) -> Problem:
  """Creates set of one-hot vectors representing a randomly generated election.

  Args:
    min_n_voters: minimum number of voters in the election.
    max_n_voters: maximum number of voters in the election.

  Returns:
    set, one-hot vector encoding the winner.
  """
  n_candidates = 20
  n_voters = random.randint(min_n_voters, max_n_voters)
  votes = np.random.randint(0, n_candidates, size=(n_voters,))
  one_hot_votes = np.eye(n_candidates)[votes]
  winner = np.argmax(np.sum(one_hot_votes, axis=0))
  one_hot_winner = np.eye(n_candidates)[winner]

  graph = jraph.GraphsTuple(
      n_node=np.asarray([n_voters]),
      n_edge=np.asarray([0]),
      nodes=one_hot_votes,
      edges=None,
      globals=np.zeros((1, n_candidates)),
      # There are no edges in our graph.
      senders=np.array([], dtype=np.int32),
      receivers=np.array([], dtype=np.int32))

  # In order to jit compile our code, we have to pad the nodes and edges of
  # the GraphsTuple to a static shape.
  graph = jraph.pad_with_graphs(graph, max_n_voters+1, 0)

  return Problem(graph=graph, labels=one_hot_winner)


def network_definition(
    graph: jraph.GraphsTuple,
    num_message_passing_steps: int = 1) -> jraph.ArrayTree:
  """Defines a graph neural network.

  Args:
    graph: Graphstuple the network processes.
    num_message_passing_steps: number of message passing steps.

  Returns:
    globals.
  """

  @jax.vmap
  def update_fn(*args):
    size = args[0].shape[-1]
    return hk.nets.MLP([size, size])(jnp.concatenate(args, axis=-1))

  for _ in range(num_message_passing_steps):
    gn = jraph.DeepSets(
        update_node_fn=update_fn,
        update_global_fn=update_fn,
        aggregate_nodes_for_globals_fn=jraph.segment_mean,
        )
    graph = gn(graph)

  return hk.Linear(graph.globals.shape[-1])(graph.globals)


def train(num_steps: int):
  """Trains a graph neural network on an electronic voting problem."""
  train_dataset = (2, 15)
  test_dataset = (16, 20)
  random.seed(42)

  network = hk.without_apply_rng(hk.transform(network_definition))
  problem = get_voting_problem(*train_dataset)
  params = network.init(jax.random.PRNGKey(42), problem.graph)

  @jax.jit
  def prediction_loss(params, problem):
    globals_ = network.apply(params, problem.graph)
    # We interpret the globals as logits for the winner.
    # Only the first graph is real, the second graph is for padding.
    log_prob = jax.nn.log_softmax(globals_[0]) * problem.labels
    return -jnp.sum(log_prob)

  @jax.jit
  def accuracy_loss(params, problem):
    globals_ = network.apply(params, problem.graph)
    # We interpret the globals as logits for the winner.
    # Only the first graph is real, the second graph is for padding.
    equal = jnp.argmax(globals_[0]) == jnp.argmax(problem.labels)
    return equal.astype(np.int32)

  opt_init, opt_update = optax.adam(2e-4)
  opt_state = opt_init(params)

  @jax.jit
  def update(params, opt_state, problem):
    g = jax.grad(prediction_loss)(params, problem)
    updates, opt_state = opt_update(g, opt_state)
    return optax.apply_updates(params, updates), opt_state

  for step in range(num_steps):
    problem = get_voting_problem(*train_dataset)
    params, opt_state = update(params, opt_state, problem)
    if step % 1000 == 0:
      train_loss = jnp.mean(
          jnp.asarray([
              accuracy_loss(params, get_voting_problem(*train_dataset))
              for _ in range(100)
          ])).item()
      test_loss = jnp.mean(
          jnp.asarray([
              accuracy_loss(params, get_voting_problem(*test_dataset))
              for _ in range(100)
          ])).item()
      logging.info("step %r loss train %r test %r", step, train_loss, test_loss)


def main(argv):
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")

  train(num_steps=100000)


if __name__ == "__main__":
  app.run(main)



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""2-SAT solver example.

Here we train a graph neural network to solve 2-sat problems.
https://en.wikipedia.org/wiki/2-satisfiability

For instance a 2 sat problem with 3 literals would look like this:
   (a or b)  and  (not a or c)  and (not b or not c)

We represent this problem in form of a bipartite-graph, with edges
connecting the literal-nodes (a, b, c) with the constraint-nodes (O).
The corresponding graph looks like this:
     O    O   O
     |\  /\  /|
     | \/  \/ |
     | /\  /\ |
     |/  \/  \|
     a    b   c

The nodes are one-hot encoded with literal nodes as (1, 0) and constraint nodes
as (0, 1). The edges are one-hot encoded with (1, 0) if the literal should be
true and (0, 1) if the literal should be false.

The graph neural network encodes the nodes and the edges and runs multiple
message passing steps by calculating message for each edge and aggregating
all the messages of the nodes.

The training dataset consists of randomly generated 2-sat problems with 2 to 15
literals.
The test dataset consists of randomly generated 2-sat problems with 16 to 20
literals.
"""

import collections
import logging
import random

from absl import app
import haiku as hk
import jax
import jax.numpy as jnp
import jraph
import numpy as np
import optax


Problem = collections.namedtuple("Problem", ("graph", "labels", "mask"))


def get_2sat_problem(min_n_literals: int, max_n_literals: int) -> Problem:
  """Creates bipartite-graph representing a randomly generated 2-sat problem.

  Args:
    min_n_literals: minimum number of literals in the 2-sat problem.
    max_n_literals: maximum number of literals in the 2-sat problem.

  Returns:
    bipartite-graph, node labels and node mask.
  """
  n_literals = random.randint(min_n_literals, max_n_literals)
  n_literals_true = random.randint(1, n_literals - 1)
  n_constraints = n_literals * (n_literals - 1) // 2

  n_node = n_literals +  n_constraints
  # 0 indicates a literal node
  # 1 indicates a constraint node.
  nodes = [0 if i < n_literals else 1 for i in range(n_node)]
  edges = []
  senders = []
  for literal_node1 in range(n_literals):
    for literal_node2 in range(literal_node1 + 1, n_literals):
      senders.append(literal_node1)
      senders.append(literal_node2)
      # 1 indicates that the literal must be true for this constraint.
      # 0 indicates that the literal must be false for this constraint.
      # I.e. with literals a and b, we have the following possible constraints:
      # 0, 0 -> a or b
      # 1, 0 -> not a or b
      # 0, 1 -> a or not b
      # 1, 1 -> not a or not b
      edges.append(1 if literal_node1 < n_literals_true else 0)
      edges.append(1 if literal_node2 < n_literals_true else 0)

  graph = jraph.GraphsTuple(
      n_node=np.asarray([n_node]),
      n_edge=np.asarray([2 * n_constraints]),
      # One-hot encoding for nodes and edges.
      nodes=np.eye(2)[nodes],
      edges=np.eye(2)[edges],
      globals=None,
      senders=np.asarray(senders),
      receivers=np.repeat(np.arange(n_constraints) + n_literals, 2))

  # In order to jit compile our code, we have to pad the nodes and edges of
  # the GraphsTuple to a static shape.
  max_n_constraints = max_n_literals * (max_n_literals - 1) // 2
  max_nodes = max_n_literals + max_n_constraints  + 1
  max_edges = 2 * max_n_constraints
  graph = jraph.pad_with_graphs(graph, max_nodes, max_edges)

  # The ground truth solution for the 2-sat problem.
  labels = (np.arange(max_nodes) < n_literals_true).astype(np.int32)
  labels = np.eye(2)[labels]

  # For the loss calculation we create a mask for the nodes, which masks the
  # the constraint nodes and the padding nodes.
  mask = (np.arange(max_nodes) < n_literals).astype(np.int32)
  return Problem(graph=graph, labels=labels, mask=mask)


def network_definition(
    graph: jraph.GraphsTuple,
    num_message_passing_steps: int = 5) -> jraph.ArrayTree:
  """Defines a graph neural network.

  Args:
    graph: Graphstuple the network processes.
    num_message_passing_steps: number of message passing steps.

  Returns:
    Decoded nodes.
  """
  embedding = jraph.GraphMapFeatures(
      embed_edge_fn=jax.vmap(hk.Linear(output_size=16)),
      embed_node_fn=jax.vmap(hk.Linear(output_size=16)))
  graph = embedding(graph)

  @jax.vmap
  @jraph.concatenated_args
  def update_fn(features):
    net = hk.Sequential([
        hk.Linear(10), jax.nn.relu,
        hk.Linear(10), jax.nn.relu,
        hk.Linear(10), jax.nn.relu])
    return net(features)

  for _ in range(num_message_passing_steps):
    gn = jraph.InteractionNetwork(
        update_edge_fn=update_fn,
        update_node_fn=update_fn,
        include_sent_messages_in_node_update=True)
    graph = gn(graph)

  return hk.Linear(2)(graph.nodes)


def train(num_steps: int):
  """Trains a graph neural network on a 2-sat problem."""
  train_dataset = (2, 15)
  test_dataset = (16, 20)
  random.seed(42)

  network = hk.without_apply_rng(hk.transform(network_definition))
  problem = get_2sat_problem(*train_dataset)
  params = network.init(jax.random.PRNGKey(42), problem.graph)

  @jax.jit
  def prediction_loss(params, problem):
    decoded_nodes = network.apply(params, problem.graph)
    # We interpret the decoded nodes as a pair of logits for each node.
    log_prob = jax.nn.log_softmax(decoded_nodes) * problem.labels
    return -jnp.sum(log_prob * problem.mask[:, None]) / jnp.sum(problem.mask)

  opt_init, opt_update = optax.adam(2e-4)
  opt_state = opt_init(params)

  @jax.jit
  def update(params, opt_state, problem):
    g = jax.grad(prediction_loss)(params, problem)
    updates, opt_state = opt_update(g, opt_state)
    return optax.apply_updates(params, updates), opt_state

  for step in range(num_steps):
    problem = get_2sat_problem(*train_dataset)
    params, opt_state = update(params, opt_state, problem)
    if step % 1000 == 0:
      train_loss = jnp.mean(
          jnp.asarray([
              prediction_loss(params, get_2sat_problem(*train_dataset))
              for _ in range(100)
          ])).item()
      test_loss = jnp.mean(
          jnp.asarray([
              prediction_loss(params, get_2sat_problem(*test_dataset))
              for _ in range(100)
          ])).item()
      logging.info("step %r loss train %r test %r", step, train_loss, test_loss)


def main(argv):
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")

  train(num_steps=10000)


if __name__ == "__main__":
  app.run(main)
